# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cycle.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0b\x63ycle.proto\x12\x04test\";\n\nLinkedList\x12\r\n\x05value\x18\x01 \x01(\x05\x12\x1e\n\x04next\x18\n \x01(\x0b\x32\x10.test.LinkedListB\tZ\x07../testb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'cycle_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\007../test'
  _LINKEDLIST._serialized_start=21
  _LINKEDLIST._serialized_end=80
# @@protoc_insertion_point(module_scope)



import sys
import logging

from confluent_kafka.admin import AdminClient

if len(sys.argv) != 2:
    sys.stderr.write("Usage: %s <broker>\n" % sys.argv[0])
    sys.exit(1)

broker = sys.argv[1]

# Custom logger
logger = logging.getLogger('AdminClient')
logger.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)-15s %(levelname)-8s %(message)s'))
logger.addHandler(handler)

# Create Admin client with logger
a = AdminClient({'bootstrap.servers': broker,
                 'debug': 'all'},
                logger=logger)

# Alternatively, pass the logger as a key.
# When passing it as an argument, it overwrites the key.
#
# a = AdminClient({'bootstrap.servers': broker,
#                  'debug': 'all',
#                  'logger': logger})

# Sample Admin API call
future = a.list_consumer_groups(request_timeout=10)

while not future.done():
    # Log messages through custom logger while waiting for the result
    a.poll(0.1)

try:
    list_consumer_groups_result = future.result()
    print("\n\n\n========================= List consumer groups result Start =========================")
    print("{} consumer groups".format(len(list_consumer_groups_result.valid)))
    for valid in list_consumer_groups_result.valid:
        print("    id: {} is_simple: {} state: {}".format(
            valid.group_id, valid.is_simple_consumer_group, valid.state))
    print("{} errors".format(len(list_consumer_groups_result.errors)))
    for error in list_consumer_groups_result.errors:
        print("    error: {}".format(error))
    print("========================= List consumer groups result End =========================\n\n\n")

except Exception:
    raise

# Log final log messages
a.poll(0)



# -*- coding: utf-8 -*-
# Copyright 2022 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
import time

from confluent_kafka import ConsumerGroupTopicPartitions, TopicPartition, ConsumerGroupState, KafkaError
from confluent_kafka.admin import (NewPartitions, ConfigResource,
                                   AclBinding, AclBindingFilter, ResourceType,
                                   ResourcePatternType, AclOperation, AclPermissionType)
from confluent_kafka.error import ConsumeError

topic_prefix = "test-topic"


# Shared between producer and consumer tests and used to verify
# that consumed headers are what was actually produced.
produce_headers = [('foo1', 'bar'),
                   ('foo1', 'bar2'),
                   ('foo2', b'1'),
                   (u'Jämtland', u'Härjedalen'),  # automatically utf-8 encoded
                   ('nullheader', None),
                   ('empty', ''),
                   ('foobin', struct.pack('hhl', 10, 20, 30))]


def verify_commit_result(err, partitions):
    assert err is not None


def verify_admin_acls(admin_client,
                      topic,
                      group):

    #
    # Add three ACLs
    #
    acl_binding_1 = AclBinding(ResourceType.TOPIC, topic, ResourcePatternType.LITERAL,
                               "User:test-user-1", "*", AclOperation.READ, AclPermissionType.ALLOW)
    acl_binding_2 = AclBinding(ResourceType.TOPIC, topic, ResourcePatternType.PREFIXED,
                               "User:test-user-2", "*", AclOperation.WRITE, AclPermissionType.DENY)
    acl_binding_3 = AclBinding(ResourceType.GROUP, group, ResourcePatternType.PREFIXED,
                               "User:test-user-2", "*", AclOperation.ALL, AclPermissionType.ALLOW)

    fs = admin_client.create_acls([acl_binding_1, acl_binding_2, acl_binding_3])
    for acl_binding, f in fs.items():
        f.result()  # trigger exception if there was an error

    time.sleep(1)

    acl_binding_filter1 = AclBindingFilter(ResourceType.ANY, None, ResourcePatternType.ANY,
                                           None, None, AclOperation.ANY, AclPermissionType.ANY)
    acl_binding_filter2 = AclBindingFilter(ResourceType.ANY, None, ResourcePatternType.PREFIXED,
                                           None, None, AclOperation.ANY, AclPermissionType.ANY)
    acl_binding_filter3 = AclBindingFilter(ResourceType.TOPIC, None, ResourcePatternType.ANY,
                                           None, None, AclOperation.ANY, AclPermissionType.ANY)
    acl_binding_filter4 = AclBindingFilter(ResourceType.GROUP, None, ResourcePatternType.ANY,
                                           None, None, AclOperation.ANY, AclPermissionType.ANY)

    expected_acl_bindings = [acl_binding_1, acl_binding_2, acl_binding_3]
    acl_bindings = admin_client.describe_acls(acl_binding_filter1).result()
    assert sorted(acl_bindings) == sorted(expected_acl_bindings), \
        "ACL bindings don't match, actual: {} expected: {}".format(acl_bindings,
                                                                   expected_acl_bindings)

    #
    # Delete the ACLs with PREFIXED
    #
    expected_acl_bindings = [acl_binding_2, acl_binding_3]
    fs = admin_client.delete_acls([acl_binding_filter2])
    deleted_acl_bindings = sorted(fs[acl_binding_filter2].result())
    assert deleted_acl_bindings == expected_acl_bindings, \
        "Deleted ACL bindings don't match, actual {} expected {}".format(deleted_acl_bindings,
                                                                         expected_acl_bindings)

    time.sleep(1)

    #
    # Delete the ACLs with TOPIC and GROUP
    #
    expected_acl_bindings = [[acl_binding_1], []]
    delete_acl_binding_filters = [acl_binding_filter3, acl_binding_filter4]
    fs = admin_client.delete_acls(delete_acl_binding_filters)
    for acl_binding, expected in zip(delete_acl_binding_filters, expected_acl_bindings):
        deleted_acl_bindings = sorted(fs[acl_binding].result())
        assert deleted_acl_bindings == expected, \
            "Deleted ACL bindings don't match, actual {} expected {}".format(deleted_acl_bindings,
                                                                             expected)

    time.sleep(1)

    #
    # All the ACLs should have been deleted
    #
    expected_acl_bindings = []
    acl_bindings = admin_client.describe_acls(acl_binding_filter1).result()
    assert acl_bindings == expected_acl_bindings, \
        "ACL bindings don't match, actual: {} expected: {}".format(acl_bindings,
                                                                   expected_acl_bindings)


def verify_topic_metadata(client, exp_topics, *args, **kwargs):
    """
    Verify that exp_topics (dict<topicname,partcnt>) is reported in metadata.
    Will retry and wait for some time to let changes propagate.

    Non-controller brokers may return the previous partition count for some
    time before being updated, in this case simply retry.
    """

    for retry in range(0, 3):
        do_retry = 0

        md = client.list_topics(*args, **kwargs)

        for exptopic, exppartcnt in exp_topics.items():
            if exptopic not in md.topics:
                print("Topic {} not yet reported in metadata: retrying".format(exptopic))
                do_retry += 1
                continue

            if len(md.topics[exptopic].partitions) < exppartcnt:
                print("Topic {} partition count not yet updated ({} != expected {}): retrying".format(
                    exptopic, len(md.topics[exptopic].partitions), exppartcnt))
                do_retry += 1
                continue

            assert len(md.topics[exptopic].partitions) == exppartcnt, \
                "Expected {} partitions for topic {}, not {}".format(
                    exppartcnt, exptopic, md.topics[exptopic].partitions)

        if do_retry == 0:
            return  # All topics okay.

        time.sleep(1)


def verify_consumer_group_offsets_operations(client, our_topic, group_id):

    # List Consumer Group Offsets check with just group name
    request = ConsumerGroupTopicPartitions(group_id)
    fs = client.list_consumer_group_offsets([request])
    f = fs[group_id]
    res = f.result()
    assert isinstance(res, ConsumerGroupTopicPartitions)
    assert res.group_id == group_id
    assert len(res.topic_partitions) == 2
    is_any_message_consumed = False
    for topic_partition in res.topic_partitions:
        assert topic_partition.topic == our_topic
        if topic_partition.offset > 0:
            is_any_message_consumed = True
    assert is_any_message_consumed

    # Alter Consumer Group Offsets check
    alter_group_topic_partitions = list(map(lambda topic_partition: TopicPartition(topic_partition.topic,
                                                                                   topic_partition.partition,
                                                                                   0),
                                            res.topic_partitions))
    alter_group_topic_partition_request = ConsumerGroupTopicPartitions(group_id,
                                                                       alter_group_topic_partitions)
    afs = client.alter_consumer_group_offsets([alter_group_topic_partition_request])
    af = afs[group_id]
    ares = af.result()
    assert isinstance(ares, ConsumerGroupTopicPartitions)
    assert ares.group_id == group_id
    assert len(ares.topic_partitions) == 2
    for topic_partition in ares.topic_partitions:
        assert topic_partition.topic == our_topic
        assert topic_partition.offset == 0

    # List Consumer Group Offsets check with group name and partitions
    list_group_topic_partitions = list(map(lambda topic_partition: TopicPartition(topic_partition.topic,
                                                                                  topic_partition.partition),
                                           ares.topic_partitions))
    list_group_topic_partition_request = ConsumerGroupTopicPartitions(group_id,
                                                                      list_group_topic_partitions)
    lfs = client.list_consumer_group_offsets([list_group_topic_partition_request])
    lf = lfs[group_id]
    lres = lf.result()

    assert isinstance(lres, ConsumerGroupTopicPartitions)
    assert lres.group_id == group_id
    assert len(lres.topic_partitions) == 2
    for topic_partition in lres.topic_partitions:
        assert topic_partition.topic == our_topic
        assert topic_partition.offset == 0


def test_basic_operations(kafka_cluster):
    num_partitions = 2
    topic_config = {"compression.type": "gzip"}

    #
    # First iteration: validate our_topic creation.
    # Second iteration: create topic.
    #
    for validate in (True, False):
        our_topic = kafka_cluster.create_topic_and_wait_propogation(topic_prefix,
                                                                    {
                                                                        "num_partitions": num_partitions,
                                                                        "config": topic_config,
                                                                        "replication_factor": 1,
                                                                    },
                                                                    validate_only=validate
                                                                    )

    admin_client = kafka_cluster.admin()

    #
    # Find the topic in list_topics
    #
    verify_topic_metadata(admin_client, {our_topic: num_partitions})
    verify_topic_metadata(admin_client, {our_topic: num_partitions}, topic=our_topic)
    verify_topic_metadata(admin_client, {our_topic: num_partitions}, our_topic)

    #
    # Increase the partition count
    #
    num_partitions += 3
    fs = admin_client.create_partitions([NewPartitions(our_topic,
                                                       new_total_count=num_partitions)],
                                        operation_timeout=10.0)

    for topic2, f in fs.items():
        f.result()  # trigger exception if there was an error

    #
    # Verify with list_topics.
    #
    verify_topic_metadata(admin_client, {our_topic: num_partitions})

    #
    # Verify with list_groups.
    #

    # Produce some messages
    p = kafka_cluster.producer()
    p.produce(our_topic, 'Hello Python!', headers=produce_headers)
    p.produce(our_topic, key='Just a key and headers', headers=produce_headers)
    p.flush()

    def consume_messages(group_id, num_messages=None):
        # Consume messages
        conf = {'group.id': group_id,
                'session.timeout.ms': 6000,
                'enable.auto.commit': False,
                'on_commit': verify_commit_result,
                'auto.offset.reset': 'earliest',
                'enable.partition.eof': True}
        c = kafka_cluster.consumer(conf)
        c.subscribe([our_topic])
        eof_reached = dict()
        read_messages = 0
        msg = None
        while True:
            try:
                msg = c.poll()
                if msg is None:
                    raise Exception('Got timeout from poll() without a timeout set: %s' % msg)
                # Commit offset
                c.commit(msg, asynchronous=False)
                read_messages += 1
                if num_messages is not None and read_messages == num_messages:
                    print('Read all the required messages: exiting')
                    break
            except ConsumeError as e:
                if e.code == KafkaError._PARTITION_EOF:
                    msg = e.kafka_message
                    print('Reached end of %s [%d] at offset %d' % (
                          msg.topic(), msg.partition(), msg.offset()))
                    eof_reached[(msg.topic(), msg.partition())] = True
                    if len(eof_reached) == len(c.assignment()):
                        print('EOF reached for all assigned partitions: exiting')
                        break
                else:
                    print('Consumer error: %s: ignoring' % str(e))
                    break
        c.close()

    group1 = 'test-group-1'
    group2 = 'test-group-2'
    acls_topic = our_topic + "-acls"
    acls_group = "test-group-acls"
    consume_messages(group1, 2)
    consume_messages(group2, 2)

    # list_groups without group argument
    groups = set(group.id for group in admin_client.list_groups(timeout=10))
    assert group1 in groups, "Consumer group {} not found".format(group1)
    assert group2 in groups, "Consumer group {} not found".format(group2)
    # list_groups with group argument
    groups = set(group.id for group in admin_client.list_groups(group1))
    assert group1 in groups, "Consumer group {} not found".format(group1)
    groups = set(group.id for group in admin_client.list_groups(group2))
    assert group2 in groups, "Consumer group {} not found".format(group2)

    # List Consumer Groups new API test
    future = admin_client.list_consumer_groups(request_timeout=10)
    result = future.result()
    group_ids = [group.group_id for group in result.valid]
    assert group1 in group_ids, "Consumer group {} not found".format(group1)
    assert group2 in group_ids, "Consumer group {} not found".format(group2)

    future = admin_client.list_consumer_groups(request_timeout=10, states={ConsumerGroupState.STABLE})
    result = future.result()
    assert isinstance(result.valid, list)
    assert not result.valid

    def verify_config(expconfig, configs):
        """
        Verify that the config key,values in expconfig are found
        and matches the ConfigEntry in configs.
        """
        for key, expvalue in expconfig.items():
            entry = configs.get(key, None)
            assert entry is not None, "Config {} not found in returned configs".format(key)

            assert entry.value == str(expvalue), \
                "Config {} with value {} does not match expected value {}".format(key, entry, expvalue)

    #
    # Get current topic config
    #
    resource = ConfigResource(ResourceType.TOPIC, our_topic)
    fs = admin_client.describe_configs([resource])
    configs = fs[resource].result()  # will raise exception on failure

    # Verify config matches our expectations
    verify_config(topic_config, configs)

    #
    # Now change the config.
    #
    topic_config["file.delete.delay.ms"] = 12345
    topic_config["compression.type"] = "snappy"

    for key, value in topic_config.items():
        resource.set_config(key, value)

    fs = admin_client.alter_configs([resource])
    fs[resource].result()  # will raise exception on failure

    time.sleep(1)

    #
    # Read the config back again and verify.
    #
    fs = admin_client.describe_configs([resource])
    configs = fs[resource].result()  # will raise exception on failure

    # Verify config matches our expectations
    verify_config(topic_config, configs)

    # Verify Consumer Offset Operations
    verify_consumer_group_offsets_operations(admin_client, our_topic, group1)

    # Delete groups
    fs = admin_client.delete_consumer_groups([group1, group2], request_timeout=10)
    fs[group1].result()  # will raise exception on failure
    fs[group2].result()  # will raise exception on failure

    #
    # Delete the topic
    #
    fs = admin_client.delete_topics([our_topic])
    fs[our_topic].result()  # will raise exception on failure
    print("Topic {} marked for deletion".format(our_topic))

    # Verify ACL operations
    verify_admin_acls(admin_client, acls_topic, acls_group)



# -*- coding: utf-8 -*-
# Copyright 2023 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from confluent_kafka import TopicPartition, IsolationLevel
from confluent_kafka.admin import ListOffsetsResultInfo, OffsetSpec


def test_list_offsets(kafka_cluster):
    """
    Test list offsets results when asking for the earliest offset (the first one),
    the latest offset (last one), max timestamp (second one)
    or after a specific timestamp (second one).
    """

    admin_client = kafka_cluster.admin()

    # Create a topic with a single partition
    topic = kafka_cluster.create_topic_and_wait_propogation("test-topic-verify-list-offsets",
                                                            {
                                                                "num_partitions": 1,
                                                                "replication_factor": 1,
                                                            })

    # Create Producer instance
    p = kafka_cluster.producer()
    base_timestamp = 1000000000
    p.produce(topic, "Message-1", timestamp=(base_timestamp + 100))
    p.produce(topic, "Message-2", timestamp=(base_timestamp + 400))
    p.produce(topic, "Message-3", timestamp=(base_timestamp + 200))
    p.flush()

    topic_partition = TopicPartition(topic, 0)

    for kwargs in [{},
                   {"isolation_level": IsolationLevel.READ_UNCOMMITTED},
                   {"request_timeout": 30},
                   {"isolation_level": IsolationLevel.READ_COMMITTED, "request_timeout": 30}]:

        requests = {topic_partition: OffsetSpec.earliest()}
        futmap = admin_client.list_offsets(requests, **kwargs)
        for _, fut in futmap.items():
            result = fut.result()
            assert isinstance(result, ListOffsetsResultInfo)
            assert (result.offset == 0)

        requests = {topic_partition: OffsetSpec.latest()}
        futmap = admin_client.list_offsets(requests, **kwargs)
        for _, fut in futmap.items():
            result = fut.result()
            assert isinstance(result, ListOffsetsResultInfo)
            assert (result.offset == 3)

        requests = {topic_partition: OffsetSpec.max_timestamp()}
        futmap = admin_client.list_offsets(requests, **kwargs)
        for _, fut in futmap.items():
            result = fut.result()
            assert isinstance(result, ListOffsetsResultInfo)
            assert (result.offset == 1)

        requests = {topic_partition: OffsetSpec.for_timestamp(base_timestamp + 150)}
        futmap = admin_client.list_offsets(requests, **kwargs)
        for _, fut in futmap.items():
            result = fut.result()
            assert isinstance(result, ListOffsetsResultInfo)
            assert (result.offset == 1)

    # Delete created topic
    fs = admin_client.delete_topics([topic])
    for topic, f in fs.items():
        f.result()



#!/usr/bin/env python3
import os
import subprocess

PYTHON_SOAK_TEST_BRANCH = 'master'

LIBRDKAFKA_VERSIONS = [
    '2.12.1',
    '2.11.1',
    '2.10.1',
    '2.8.0',
    '2.6.1',
    '2.5.3',
    '2.4.0',
    '2.3.0',
    '2.2.0',
    '2.1.1',
    '2.0.2'
]

PYTHON_VERSIONS = [
    '2.12.1',
    '2.11.1',
    '2.10.1',
    '2.8.0',
    '2.6.1',
    '2.5.3',
    '2.4.0',
    '2.3.0',
    '2.2.0',
    '2.1.1',
    '2.0.2'
]


def run(command):
    print(f'Running command: {command}')
    env = os.environ.copy()
    env['MAKEFLAGS'] = 'SHELL=/bin/bash -j8'  # For parallel make jobs
    result = subprocess.run(command, shell=True, env=env)
    if result.returncode != 0:
        raise Exception(f'Command failed: {command}, result: {result.returncode}')


if __name__ == '__main__':
    for i, librdkafka_version in enumerate(LIBRDKAFKA_VERSIONS):
        python_version = PYTHON_VERSIONS[i]
        librdkafka_version_folder_name = librdkafka_version.replace('.', '_')
        folder_name = f'confluent-kafka-python_{librdkafka_version_folder_name}'
        if os.path.exists(folder_name):
            print(f'Skipping \'{folder_name}\', already exists.')
            continue

        run('git clone https://github.com/confluentinc/confluent-kafka-python.git')
        run(f'mv confluent-kafka-python {folder_name}')
        run(f'cd {folder_name} && git checkout {PYTHON_SOAK_TEST_BRANCH}')
        run(f'cd {folder_name}/tests/soak && ./build.sh v{librdkafka_version} v{python_version}')
        run(f'cp config/*.config {folder_name}/tests/soak')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
from confluent_kafka import TopicPartition, OFFSET_END, KafkaError, KafkaException

from confluent_kafka.error import ConsumeError
from confluent_kafka.serialization import StringSerializer
from tests.common import TestUtils


def test_consume_error(kafka_cluster):
    """
    Tests to ensure librdkafka errors are propagated as
    an instance of ConsumeError.
    """
    topic = kafka_cluster.create_topic_and_wait_propogation("test_commit_transaction")
    consumer_conf = {'group.id': 'pytest', 'enable.partition.eof': True}

    producer = kafka_cluster.producer()
    producer.produce(topic=topic, value="a")
    producer.flush()

    consumer = kafka_cluster.consumer(consumer_conf,
                                      value_deserializer=StringSerializer())
    consumer.assign([TopicPartition(topic, 0, OFFSET_END)])

    with pytest.raises(ConsumeError) as exc_info:
        # Trigger EOF error
        consumer.poll()
    assert exc_info.value.args[0].code() == KafkaError._PARTITION_EOF, \
        "Expected _PARTITION_EOF, not {}".format(exc_info)


# Skipping the test for consumer protocol for now. Update the test to use
# IncrementalAlterConfigs Admin operation to update
# group.session.timeout.ms and enable the test again.
@pytest.mark.skipif(TestUtils.use_group_protocol_consumer(),
                    reason="session.timeout.ms is not supported on client side for "
                    "consumer protocol. Update this test to use IncrementalAlterConfigs "
                    "Admin operation to update group.session.timeout.ms and enable "
                    "the test again.")
def test_consume_error_commit(kafka_cluster):
    """
    Tests to ensure that we handle messages with errors when commiting.
    """
    topic = kafka_cluster.create_topic_and_wait_propogation("test_commit_transaction")
    consumer_conf = {'group.id': 'pytest',
                     'session.timeout.ms': 100}

    producer = kafka_cluster.producer()
    producer.produce(topic=topic, value="a")
    producer.flush()

    consumer = kafka_cluster.cimpl_consumer(consumer_conf)
    consumer.subscribe([topic])
    try:
        # Since the session timeout value is low, JoinGroupRequest will fail
        # and we get error in a message while polling.
        m = consumer.poll(2)
        consumer.commit(m)
    except KafkaException as e:
        assert e.args[0].code() == KafkaError._INVALID_ARG, \
            "Expected INVALID_ARG, not {}".format(e)


# Skipping the test for consumer protocol for now. Update the test to use
# IncrementalAlterConfigs Admin operation to update
# group.session.timeout.ms and enable the test again.
@pytest.mark.skipif(TestUtils.use_group_protocol_consumer(),
                    reason="session.timeout.ms is not supported on client side for "
                    "consumer protocol. Update this test to use IncrementalAlterConfigs "
                    "Admin operation to update group.session.timeout.ms and enable "
                    "the test again.")
def test_consume_error_store_offsets(kafka_cluster):
    """
    Tests to ensure that we handle messages with errors when storing offsets.
    """
    topic = kafka_cluster.create_topic_and_wait_propogation("test_commit_transaction")
    consumer_conf = {'group.id': 'pytest',
                     'session.timeout.ms': 100,
                     'enable.auto.offset.store': True,
                     'enable.auto.commit': False}

    producer = kafka_cluster.producer()
    producer.produce(topic=topic, value="a")
    producer.flush()

    consumer = kafka_cluster.cimpl_consumer(consumer_conf)
    consumer.subscribe([topic])
    try:
        # Since the session timeout value is low, JoinGroupRequest will fail
        # and we get error in a message while polling.
        m = consumer.poll(2)
        consumer.store_offsets(m)
    except KafkaException as e:
        assert e.args[0].code() == KafkaError._INVALID_ARG, \
            "Expected INVALID_ARG, not {}".format(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2025 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from enum import Enum
from confluent_kafka import ConsumerGroupType, KafkaException
from tests.common import TestUtils

topic_prefix = "test_consumer_upgrade_downgrade_"
number_of_partitions = 10


def get_group_protocol_type(a, group_id):
    futureMap = a.describe_consumer_groups([group_id])
    try:
        future = futureMap[group_id]
        g = future.result()
        return g.type
    except KafkaException as e:
        print("Error while describing group id '{}': {}".format(group_id, e))
    except Exception:
        raise


def check_consumer(kafka_cluster, consumers, admin_client, group_id, topic, expected_protocol):
    no_of_messages = 100
    total_msg_read = 0
    expected_partitions_per_consumer = number_of_partitions // len(consumers)
    while len(consumers[-1].assignment()) != expected_partitions_per_consumer:
        for consumer in consumers:
            consumer.poll(0.1)

    all_assignments = set()
    for consumer in consumers:
        assignment = consumer.assignment()
        all_assignments.update(assignment)
        assert len(assignment) == expected_partitions_per_consumer
    assert len(all_assignments) == number_of_partitions

    assert get_group_protocol_type(admin_client, group_id) == expected_protocol

    # Produce some messages to the topic
    test_data = ['test-data{}'.format(i) for i in range(0, no_of_messages)]
    test_keys = ['test-key{}'.format(i) for i in range(0, no_of_messages)]  # we want each partition to have data
    kafka_cluster.seed_topic(topic, test_data, test_keys)

    while total_msg_read < no_of_messages:
        for consumer in consumers:
            # Poll for messages
            msg = consumer.poll(0.1)
            if msg is not None:
                total_msg_read += 1

    assert total_msg_read == no_of_messages, f"Expected to read {no_of_messages} messages, but read {total_msg_read}"


class Operation(Enum):
    ADD = 0
    REMOVE = 1


def perform_consumer_upgrade_downgrade_test_with_partition_assignment_strategy(
        kafka_cluster, partition_assignment_strategy):
    """
    Test consumer upgrade and downgrade.
    """
    topic_name_prefix = f"{topic_prefix}_{partition_assignment_strategy}"
    topic = kafka_cluster.create_topic_and_wait_propogation(topic_name_prefix,
                                                            {
                                                                "num_partitions": number_of_partitions
                                                            })
    admin_client = kafka_cluster.admin()

    consumer_conf = {'group.id': topic,
                     'auto.offset.reset': 'earliest'}
    consumer_conf_classic = {
        'group.protocol': 'classic',
        'partition.assignment.strategy': partition_assignment_strategy,
        **consumer_conf
    }
    consumer_conf_consumer = {
        'group.protocol': 'consumer',
        **consumer_conf
    }

    test_scenarios = [(Operation.ADD, consumer_conf_classic, ConsumerGroupType.CLASSIC),
                      (Operation.ADD, consumer_conf_consumer, ConsumerGroupType.CONSUMER),
                      (Operation.REMOVE, None, ConsumerGroupType.CONSUMER),
                      (Operation.ADD, consumer_conf_classic, ConsumerGroupType.CONSUMER),
                      (Operation.REMOVE, None, ConsumerGroupType.CLASSIC)]
    consumers = []

    for operation, conf, expected_protocol in test_scenarios:
        if operation == Operation.ADD:
            consumer = kafka_cluster.consumer(conf)
            assert consumer is not None
            consumer.subscribe([topic])
            consumers.append(consumer)
        elif operation == Operation.REMOVE:
            consumer_to_remove = consumers.pop(0)
            consumer_to_remove.close()
        check_consumer(kafka_cluster, consumers, admin_client, topic, topic, expected_protocol)

    assert len(consumers) == 1
    consumers[0].close()
    kafka_cluster.delete_topic(topic)


@pytest.mark.skipif(not TestUtils.use_group_protocol_consumer(),
                    reason="Skipping test as group protocol consumer is not enabled")
def test_consumer_upgrade_downgrade(kafka_cluster):
    perform_consumer_upgrade_downgrade_test_with_partition_assignment_strategy(kafka_cluster, 'roundrobin')
    perform_consumer_upgrade_downgrade_test_with_partition_assignment_strategy(kafka_cluster, 'range')
    perform_consumer_upgrade_downgrade_test_with_partition_assignment_strategy(kafka_cluster, 'cooperative-sticky')



"""
Consumer strategies for testing sync and async Kafka consumers.

This module contains strategy classes that encapsulate the different consumer
implementations (sync vs async) with consistent interfaces for testing.
"""

import time
import asyncio
import json
import os
from confluent_kafka import Consumer
from confluent_kafka.experimental.aio import AIOConsumer
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry._sync.json_schema import JSONDeserializer
from confluent_kafka.schema_registry._sync.protobuf import ProtobufDeserializer
from confluent_kafka.schema_registry.avro import AvroDeserializer
from confluent_kafka.schema_registry._async.json_schema import AsyncJSONDeserializer
from confluent_kafka.schema_registry._async.protobuf import AsyncProtobufDeserializer
from confluent_kafka.schema_registry._async.avro import AsyncAvroDeserializer
from confluent_kafka.schema_registry import AsyncSchemaRegistryClient
from confluent_kafka.serialization import (
    StringDeserializer,
    SerializationContext,
    MessageField,
)
from tests.integration.schema_registry.data.proto import PublicTestProto_pb2


class DeserializedMessage:
    """Wrapper for messages with deserialized key/value"""

    def __init__(self, original_msg, key, value):
        self._original_msg = original_msg
        self._key = key
        self._value = value

    def key(self):
        return self._key

    def value(self):
        return self._value

    def error(self):
        return self._original_msg.error()

    def topic(self):
        return self._original_msg.topic()

    def partition(self):
        return self._original_msg.partition()

    def offset(self):
        return self._original_msg.offset()


class ConsumerStrategy:
    """Base class for consumer strategies"""

    def __init__(self, bootstrap_servers, group_id, logger, batch_size=10):
        self.bootstrap_servers = bootstrap_servers
        self.group_id = group_id
        self.logger = logger
        self.batch_size = batch_size
        self.metrics = None

    def create_consumer(self):
        raise NotImplementedError()

    def consume_messages(
        self,
        topic_name,
        test_duration,
        start_time,
        consumed_container,
        timeout=1.0,
        serialization_type=None,
    ):
        raise NotImplementedError()

    def _get_base_config(self):
        """Get shared Kafka consumer configuration optimized for low-latency, high-throughput"""
        return {
            "bootstrap.servers": self.bootstrap_servers,
            "group.id": self.group_id,
            "auto.offset.reset": "earliest",
            "enable.auto.commit": "true",
            "auto.commit.interval.ms": "5000",
        }

    def _log_configuration(self, config, consumer_type, extra_params=None):
        """Log consumer configuration for validation"""
        separator = "=" * (len(f"{consumer_type.upper()} CONSUMER CONFIGURATION") + 6)
        self.logger.info(f"=== {consumer_type.upper()} CONSUMER CONFIGURATION ===")
        for key, value in config.items():
            self.logger.info(f"{key}: {value}")

        if extra_params:
            for key, value in extra_params.items():
                self.logger.info(f"{key}: {value}")

        self.logger.info(separator)

    def _get_schema_registry_client(self, is_async=False):
        """Get Schema Registry client with proper configuration"""
        schema_registry_url = os.getenv(
            "SCHEMA_REGISTRY_URL",
            getattr(self, "schema_registry_url", "http://localhost:8081"),
        )
        client_config = {
            "url": schema_registry_url,
        }

        if is_async:
            return AsyncSchemaRegistryClient(client_config)
        else:
            return SchemaRegistryClient(client_config)

    def _get_schemas(self, deserialization_type):
        """Get schema definitions for the given deserialization type, for testing purposes"""
        if deserialization_type == "avro":
            # Match the Protobuf TestMessage structure
            return {
                "type": "record",
                "name": "TestMessage",
                "fields": [
                    {"name": "test_string", "type": "string"},
                    {"name": "test_bool", "type": "boolean"},
                    {"name": "test_bytes", "type": "bytes"},
                    {"name": "test_double", "type": "double"},
                    {"name": "test_float", "type": "float"},
                    {"name": "test_fixed32", "type": "int"},
                    {"name": "test_fixed64", "type": "long"},
                    {"name": "test_int32", "type": "int"},
                    {"name": "test_int64", "type": "long"},
                    {"name": "test_sfixed32", "type": "int"},
                    {"name": "test_sfixed64", "type": "long"},
                    {"name": "test_sint32", "type": "int"},
                    {"name": "test_sint64", "type": "long"},
                    {"name": "test_uint32", "type": "int"},
                    {"name": "test_uint64", "type": "long"},
                ],
            }
        elif deserialization_type == "json":
            # Match the Protobuf TestMessage structure
            return {
                "type": "object",
                "properties": {
                    "test_string": {"type": "string"},
                    "test_bool": {"type": "boolean"},
                    "test_bytes": {"type": "string"},
                    "test_double": {"type": "number"},
                    "test_float": {"type": "number"},
                    "test_fixed32": {"type": "integer"},
                    "test_fixed64": {"type": "integer"},
                    "test_int32": {"type": "integer"},
                    "test_int64": {"type": "integer"},
                    "test_sfixed32": {"type": "integer"},
                    "test_sfixed64": {"type": "integer"},
                    "test_sint32": {"type": "integer"},
                    "test_sint64": {"type": "integer"},
                    "test_uint32": {"type": "integer"},
                    "test_uint64": {"type": "integer"},
                },
                "required": [
                    "test_string",
                    "test_bool",
                    "test_bytes",
                    "test_double",
                    "test_float",
                    "test_fixed32",
                    "test_fixed64",
                    "test_int32",
                    "test_int64",
                    "test_sfixed32",
                    "test_sfixed64",
                    "test_sint32",
                    "test_sint64",
                    "test_uint32",
                    "test_uint64",
                ],
            }
        elif deserialization_type == "protobuf":
            return PublicTestProto_pb2.TestMessage
        else:
            raise ValueError(
                f"Unsupported deserialization type: {deserialization_type}"
            )

    def _record_message_metrics(self, msg, latency_ms):
        """Shared metrics recording logic"""
        if self.metrics:
            # For deserialized messages, use original raw bytes for size calculation
            if hasattr(msg, "_original_msg"):
                # DeserializedMessage - use original raw bytes
                original_msg = msg._original_msg
                value_size = len(original_msg.value()) if original_msg.value() else 0
                key_size = len(original_msg.key()) if original_msg.key() else 0
            else:
                # Raw message - use directly
                value_size = len(msg.value()) if msg.value() else 0
                key_size = len(msg.key()) if msg.key() else 0

            message_size = value_size + key_size
            self.metrics.record_processed_message(
                message_size=message_size,
                topic=msg.topic(),
                partition=msg.partition(),
                offset=msg.offset(),
                operation_latency_ms=latency_ms,
            )


class SyncConsumerStrategy(ConsumerStrategy):
    def build_deserializers(self, deserialization_type):
        """Build sync deserializers"""
        if not deserialization_type:
            return None, None

        key_deserializer = StringDeserializer("utf8")
        sr_client = self._get_schema_registry_client(is_async=False)
        schema = self._get_schemas(deserialization_type)

        if deserialization_type == "avro":
            value_deserializer = AvroDeserializer(
                schema_registry_client=sr_client, schema_str=json.dumps(schema)
            )
        elif deserialization_type == "json":
            value_deserializer = JSONDeserializer(json.dumps(schema))
        elif deserialization_type == "protobuf":
            value_deserializer = ProtobufDeserializer(schema)

        return key_deserializer, value_deserializer

    def _deserialize_message(
        self, msg, key_deserializer, value_deserializer, topic_name, message_count
    ):
        """Sync deserialization logic"""
        if not (key_deserializer or value_deserializer):
            return msg

        try:
            # Deserialize key and value
            deserialized_key = msg.key()
            if key_deserializer and msg.key() is not None:
                deserialized_key = key_deserializer(msg.key())

            deserialized_value = msg.value()
            if value_deserializer and msg.value() is not None:
                deserialized_value = value_deserializer(
                    msg.value(), SerializationContext(topic_name, MessageField.VALUE)
                )

            # Log successful deserialization for first few messages
            if message_count < 5:
                self.logger.debug(
                    f"Deserialized message {message_count}: "
                    f"key={deserialized_key}, value={deserialized_value}"
                )

            return DeserializedMessage(msg, deserialized_key, deserialized_value)

        except Exception as e:
            self.logger.error(f"Deserialization error: {e}")
            return msg  # Return original message if deserialization fails

    def create_consumer(self, config_overrides=None):
        config = self._get_base_config()

        # Apply any test-specific overrides
        if config_overrides:
            config.update(config_overrides)

        consumer = Consumer(config)

        # Log the configuration for validation
        self._log_configuration(config, "SYNC")

        return consumer

    def consume_messages(
        self,
        topic_name,
        test_duration,
        start_time,
        consumed_container,
        timeout=1.0,
        serialization_type=None,
    ):
        # Initialize deserializers if using Schema Registry
        key_deserializer, value_deserializer = self.build_deserializers(
            serialization_type
        )

        consumer = self.create_consumer()

        try:
            # Sync consume implementation
            messages_consumed = 0

            consumer.subscribe([topic_name])

            while time.time() - start_time < test_duration:
                consume_start = time.time()
                messages = consumer.consume(
                    num_messages=self.batch_size, timeout=timeout
                )
                consume_end = time.time()

                consume_latency_ms = (consume_end - consume_start) * 1000

                if self.metrics:
                    self.metrics.record_api_call(consume_latency_ms)
                    self.metrics.record_batch_operation(
                        len(messages) if messages else 0
                    )

                if not messages:
                    if self.metrics:
                        self.metrics.record_timeout()
                    continue

                # Process all messages in the batch
                batch_consumed = 0
                for msg in messages:
                    if msg.error():
                        if self.metrics:
                            self.metrics.record_error(str(msg.error()))
                        self.logger.error(f"Consumer error: {msg.error()}")
                        continue

                    # Deserialize message
                    msg = self._deserialize_message(
                        msg,
                        key_deserializer,
                        value_deserializer,
                        topic_name,
                        messages_consumed + batch_consumed,
                    )
                    consumed_container.append(msg)
                    messages_consumed += 1
                    batch_consumed += 1
                    self._record_message_metrics(
                        msg, consume_latency_ms / max(len(messages), 1)
                    )

        finally:
            consumer.close()
        return messages_consumed

    def poll_messages(
        self,
        topic_name,
        test_duration,
        start_time,
        consumed_container,
        timeout=1.0,
        serialization_type=None,
    ):
        # Initialize deserializers if using Schema Registry
        key_deserializer, value_deserializer = self.build_deserializers(
            serialization_type
        )

        consumer = self.create_consumer()

        try:
            # Sync poll implementation
            messages_consumed = 0
            consumer.subscribe([topic_name])

            while time.time() - start_time < test_duration:
                poll_start = time.time()
                msg = consumer.poll(timeout=timeout)
                poll_end = time.time()

                poll_latency_ms = (poll_end - poll_start) * 1000

                if self.metrics:
                    self.metrics.record_api_call(poll_latency_ms)

                if msg is None:
                    if self.metrics:
                        self.metrics.record_timeout()
                    continue

                if msg.error():
                    if self.metrics:
                        self.metrics.record_error(str(msg.error()))
                    self.logger.error(f"Consumer error: {msg.error()}")
                    continue

                # Deserialize message
                msg = self._deserialize_message(
                    msg,
                    key_deserializer,
                    value_deserializer,
                    topic_name,
                    messages_consumed,
                )
                consumed_container.append(msg)
                messages_consumed += 1
                self._record_message_metrics(msg, poll_latency_ms)
        finally:
            consumer.close()
        return messages_consumed


class AsyncConsumerStrategy(ConsumerStrategy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._consumer_instance = None

    async def build_deserializers(self, deserialization_type):
        """Build async deserializers - must be called from async context"""
        if not deserialization_type:
            return None, None

        key_deserializer = StringDeserializer("utf8")
        sr_client = self._get_schema_registry_client(is_async=True)
        schema = self._get_schemas(deserialization_type)

        if deserialization_type == "avro":
            value_deserializer = await AsyncAvroDeserializer(
                schema_registry_client=sr_client, schema_str=json.dumps(schema)
            )
        elif deserialization_type == "json":
            value_deserializer = await AsyncJSONDeserializer(json.dumps(schema))
        elif deserialization_type == "protobuf":
            value_deserializer = await AsyncProtobufDeserializer(schema)

        return key_deserializer, value_deserializer

    async def _deserialize_message(
        self, msg, key_deserializer, value_deserializer, topic_name, message_count
    ):
        """Async deserialization logic"""
        if not (key_deserializer or value_deserializer):
            return msg

        try:
            # Deserialize key and value
            deserialized_key = msg.key()
            if key_deserializer and msg.key() is not None:
                # Note: StringDeserializer is sync, so no await needed
                deserialized_key = key_deserializer(msg.key())

            deserialized_value = msg.value()
            if value_deserializer and msg.value() is not None:
                deserialized_value = await value_deserializer(
                    msg.value(), SerializationContext(topic_name, MessageField.VALUE)
                )

            # Log successful deserialization for first few messages
            if message_count < 5:
                self.logger.debug(
                    f"Async deserialized message {message_count}: "
                    f"key={deserialized_key}, value={deserialized_value}"
                )

            return DeserializedMessage(msg, deserialized_key, deserialized_value)

        except Exception as e:
            self.logger.error(f"Async deserialization error: {e}")
            return msg  # Return original message if deserialization fails

    def create_consumer(self, config_overrides=None):
        config = self._get_base_config()

        # Apply any test-specific overrides
        if config_overrides:
            config.update(config_overrides)

        max_workers = getattr(self, "max_workers", 20)

        self._consumer_instance = AIOConsumer(config, max_workers=max_workers)

        # Log the configuration for validation
        extra_params = {"max_workers": max_workers}
        self._log_configuration(config, "ASYNC", extra_params)

        return self._consumer_instance

    def consume_messages(
        self,
        topic_name,
        test_duration,
        start_time,
        consumed_container,
        timeout=1.0,
        serialization_type=None,
    ):
        async def async_consume():
            # Initialize deserializers if using Schema Registry
            key_deserializer, value_deserializer = await self.build_deserializers(
                serialization_type
            )

            config_overrides = getattr(self, "config_overrides", None)
            consumer = self.create_consumer(config_overrides)

            try:
                # Async consume implementation
                messages_consumed = 0

                await consumer.subscribe([topic_name])

                while time.time() - start_time < test_duration:
                    consume_start = time.time()
                    messages = await consumer.consume(
                        num_messages=self.batch_size, timeout=timeout
                    )
                    consume_end = time.time()

                    consume_latency_ms = (consume_end - consume_start) * 1000

                    if self.metrics:
                        self.metrics.record_api_call(consume_latency_ms)
                        self.metrics.record_batch_operation(
                            len(messages) if messages else 0
                        )

                    if not messages:
                        if self.metrics:
                            self.metrics.record_timeout()
                        continue

                    # Process all messages in the batch
                    batch_consumed = 0
                    for msg in messages:
                        if msg.error():
                            if self.metrics:
                                self.metrics.record_error(str(msg.error()))
                            self.logger.error(f"Consumer error: {msg.error()}")
                            continue

                        # Deserialize message
                        msg = await self._deserialize_message(
                            msg,
                            key_deserializer,
                            value_deserializer,
                            topic_name,
                            messages_consumed + batch_consumed,
                        )
                        consumed_container.append(msg)
                        messages_consumed += 1
                        batch_consumed += 1
                        self._record_message_metrics(
                            msg, consume_latency_ms / max(len(messages), 1)
                        )

            finally:
                await consumer.close()
            return messages_consumed

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(async_consume())

    def poll_messages(
        self,
        topic_name,
        test_duration,
        start_time,
        consumed_container,
        timeout=1.0,
        serialization_type=None,
    ):
        async def async_poll():
            # Initialize deserializers if using Schema Registry
            key_deserializer, value_deserializer = await self.build_deserializers(
                serialization_type
            )

            config_overrides = getattr(self, "config_overrides", None)
            consumer = self.create_consumer(config_overrides)

            try:
                # Async poll implementation
                messages_consumed = 0
                await consumer.subscribe([topic_name])

                while time.time() - start_time < test_duration:
                    poll_start = time.time()
                    msg = await consumer.poll(timeout=timeout)
                    poll_end = time.time()

                    poll_latency_ms = (poll_end - poll_start) * 1000

                    if self.metrics:
                        self.metrics.record_api_call(poll_latency_ms)

                    if msg is None:
                        if self.metrics:
                            self.metrics.record_timeout()
                        continue

                    if msg.error():
                        if self.metrics:
                            self.metrics.record_error(str(msg.error()))
                        self.logger.error(f"Consumer error: {msg.error()}")
                        continue

                    # Deserialize message
                    msg = await self._deserialize_message(
                        msg,
                        key_deserializer,
                        value_deserializer,
                        topic_name,
                        messages_consumed,
                    )
                    consumed_container.append(msg)
                    messages_consumed += 1
                    self._record_message_metrics(msg, poll_latency_ms)

            finally:
                await consumer.close()
            return messages_consumed

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(async_poll())



"""
Consumer benchmark metrics collection and validation for Kafka performance testing.

Implements consumer-specific metrics tracking including poll latencies,
consumption rates, per-topic breakdowns, and reliability analysis.
"""
import time
import statistics
from typing import List, Dict, Any
from collections import defaultdict
import psutil


class ConsumerMetricsCollector:
    """Collects comprehensive performance metrics for consumer testing."""

    def __init__(self, operation_type: str = "poll", serialization_type: str = None):
        # Basic timing
        self.start_time = None
        self.end_time = None

        # Operation type: "poll" or "consume"
        self.operation_type = operation_type
        self.serialization_type = serialization_type

        # Consumer metrics (generic for both poll/consume)
        self.messages_consumed = 0
        self.operation_attempts = 0  # poll_attempts or consume_attempts
        self.operation_timeouts = 0  # poll_timeouts or consume_timeouts
        self.operation_errors = 0    # poll_errors or consume_errors
        self.operation_latencies = []  # in milliseconds
        self.error_messages = []

        # Data tracking
        self.total_bytes = 0
        self.message_sizes = []

        # Simple offset tracking for single topic/partition scenarios
        self.offsets_consumed = defaultdict(dict)  # topic -> partition -> last_offset

        # Memory tracking
        self.initial_memory_mb = None
        self.peak_memory_mb = 0

        # Batch efficiency tracking
        self.consume_call_count = 0
        self.empty_consume_count = 0

        # Process object for efficient memory monitoring
        self._process = None

    def start(self):
        """Start metrics collection"""
        self.start_time = time.time()
        self.initial_memory_mb = self._get_memory_usage()

    def record_api_call(self, operation_latency_ms: float):
        """Record a consumer API call (poll or consume) with its latency"""
        self.operation_attempts += 1
        self.operation_latencies.append(operation_latency_ms)

    def record_timeout(self, topic: str = "unknown"):
        """Record a consumer operation timeout"""
        self.operation_timeouts += 1

    def record_error(self, error_msg: str, topic: str = "unknown"):
        """Record a consumer error"""
        self.operation_errors += 1
        self.error_messages.append(error_msg)

    def record_processed_message(self, message_size: int, topic: str, partition: int,
                                 offset: int, operation_latency_ms: float):
        """Record a successfully processed message"""
        self.messages_consumed += 1
        self.total_bytes += message_size
        self.message_sizes.append(message_size)

        # Track offsets
        self.offsets_consumed[topic][partition] = offset

        # Update peak memory usage
        self._update_memory_usage()

    def finalize(self):
        """Finalize metrics collection"""
        self.end_time = time.time()

    def record_batch_operation(self, messages_returned: int):
        """Record a batch operation (consume() call) and how many messages it returned"""
        self.consume_call_count += 1
        if messages_returned == 0:
            self.empty_consume_count += 1

    def _get_memory_usage(self) -> float:
        """Get current memory usage in MB"""
        if psutil is None:
            return 0.0
        try:
            if self._process is None:
                self._process = psutil.Process()
            return self._process.memory_info().rss / (1024 * 1024)
        except (psutil.Error, OSError):
            return 0.0

    def _update_memory_usage(self):
        """Update peak memory usage"""
        current_memory = self._get_memory_usage()
        if current_memory > self.peak_memory_mb:
            self.peak_memory_mb = current_memory

    def _percentile(self, data, percentile):
        """Calculate percentile for datasets where quantiles() fails"""
        if not data:
            return 0
        sorted_data = sorted(data)
        k = (len(sorted_data) - 1) * percentile / 100.0
        f = int(k)
        c = k - f
        if f == len(sorted_data) - 1:
            return sorted_data[f]
        return sorted_data[f] * (1 - c) + sorted_data[f + 1] * c

    def get_summary(self) -> Dict[str, Any]:
        """Get consumer-specific metrics summary"""
        if not self.start_time or not self.end_time:
            return {}

        duration = self.end_time - self.start_time

        # Consumer-specific calculations
        consumption_rate = self.messages_consumed / duration if duration > 0 else 0
        operation_rate = self.operation_attempts / duration if duration > 0 else 0

        # Operation metrics (generic for poll/consume)
        operation_error_rate = (self.operation_errors / self.operation_attempts
                                if self.operation_attempts > 0 else 0)
        operation_success_rate = ((self.operation_attempts - self.operation_timeouts -
                                   self.operation_errors) / self.operation_attempts
                                  if self.operation_attempts > 0 else 0)

        # Operation latency analysis
        if self.operation_latencies:
            avg_operation_latency = statistics.mean(self.operation_latencies)
            p50_operation_latency = statistics.median(self.operation_latencies)
            max_operation_latency = max(self.operation_latencies)

            try:
                quantiles = statistics.quantiles(self.operation_latencies, n=100)
                p95_operation_latency = quantiles[94]  # 95th percentile
                p99_operation_latency = quantiles[98]  # 99th percentile
            except statistics.StatisticsError:
                p95_operation_latency = self._percentile(self.operation_latencies, 95)
                p99_operation_latency = self._percentile(self.operation_latencies, 99)
        else:
            avg_operation_latency = p50_operation_latency = p95_operation_latency = (
                p99_operation_latency) = max_operation_latency = 0

        # Message size analysis
        if self.message_sizes:
            avg_message_size = statistics.mean(self.message_sizes)
            min_message_size = min(self.message_sizes)
            max_message_size = max(self.message_sizes)
            median_message_size = statistics.median(self.message_sizes)
        else:
            avg_message_size = min_message_size = max_message_size = median_message_size = 0

        # Memory usage analysis
        memory_growth_mb = 0
        if self.initial_memory_mb and self.peak_memory_mb:
            memory_growth_mb = self.peak_memory_mb - self.initial_memory_mb

        # Batch efficiency analysis
        messages_per_consume = self.messages_consumed / self.consume_call_count if self.consume_call_count > 0 else 0
        empty_consume_rate = self.empty_consume_count / self.consume_call_count if self.consume_call_count > 0 else 0

        # Base summary
        base_summary = {
            'start_time': self.start_time,
            'end_time': self.end_time,
            'duration_seconds': duration
        }

        # Consumer metrics
        # Dynamic metric names based on operation type
        op_name = self.operation_type  # "poll" or "consume"

        consumer_metrics = {
            # Basic metrics
            'messages_consumed': self.messages_consumed,
            'consumption_rate_msg_per_sec': consumption_rate,
            'data_throughput_mb_per_sec': (self.total_bytes / (1024 * 1024)) / duration if duration > 0 else 0,
            'avg_latency_ms': avg_operation_latency,
            'p50_latency_ms': p50_operation_latency,
            'p95_latency_ms': p95_operation_latency,
            'p99_latency_ms': p99_operation_latency,
            'max_latency_ms': max_operation_latency,
            'error_rate': operation_error_rate,
            'success_rate': operation_success_rate,
            'total_bytes': self.total_bytes,

            # Enhanced metrics
            'avg_message_size_bytes': avg_message_size,
            'median_message_size_bytes': median_message_size,
            'min_message_size_bytes': min_message_size,
            'max_message_size_bytes': max_message_size,
            'memory_growth_mb': memory_growth_mb,
            'peak_memory_mb': self.peak_memory_mb,

            # Operation-specific metrics (dynamic naming)
            f'{op_name}_attempts': self.operation_attempts,
            f'{op_name}_rate_per_sec': operation_rate,
            f'{op_name}_timeouts': self.operation_timeouts,
            f'{op_name}_errors': self.operation_errors,
            f'messages_per_{op_name}': messages_per_consume,  # Will be messages_per_poll or messages_per_consume
            f'empty_{op_name}_rate': empty_consume_rate,  # Will be empty_poll_rate or empty_consume_rate
            f'{op_name}_call_count': self.consume_call_count,  # Generic call count
            f'empty_{op_name}_count': self.empty_consume_count,
        }

        base_summary.update(consumer_metrics)
        return base_summary


class ConsumerMetricsBounds:
    """Performance bounds for consumer metrics validation"""

    def __init__(self,
                 min_consumption_rate: float = 1.0,
                 max_avg_latency_ms: float = 5000.0,
                 max_p95_latency_ms: float = 10000.0,
                 min_success_rate: float = 0.90,
                 max_error_rate: float = 0.05,
                 max_memory_growth_mb: float = 600.0,
                 min_messages_per_consume: float = 0.5,
                 max_empty_consume_rate: float = 0.5):
        self.min_consumption_rate = min_consumption_rate
        self.max_avg_latency_ms = max_avg_latency_ms
        self.max_p95_latency_ms = max_p95_latency_ms
        self.min_success_rate = min_success_rate
        self.max_error_rate = max_error_rate
        # Enhanced bounds
        self.max_memory_growth_mb = max_memory_growth_mb
        self.min_messages_per_consume = min_messages_per_consume
        self.max_empty_consume_rate = max_empty_consume_rate


def validate_consumer_metrics(metrics: Dict[str, Any], bounds: ConsumerMetricsBounds) -> tuple[bool, List[str]]:
    """Validate consumer metrics against performance bounds"""
    violations = []

    # Consumption rate check
    consumption_rate = metrics.get('consumption_rate_msg_per_sec', 0)
    if consumption_rate < bounds.min_consumption_rate:
        violations.append(f"Consumption rate {consumption_rate:.2f} msg/s below minimum {bounds.min_consumption_rate}")

    # Latency checks
    avg_latency = metrics.get('avg_latency_ms', 0)
    if avg_latency > bounds.max_avg_latency_ms:
        violations.append(f"Average latency {avg_latency:.2f}ms exceeds maximum {bounds.max_avg_latency_ms}ms")

    p95_latency = metrics.get('p95_latency_ms', 0)
    if p95_latency > bounds.max_p95_latency_ms:
        violations.append(f"P95 latency {p95_latency:.2f}ms exceeds maximum {bounds.max_p95_latency_ms}ms")

    # Success rate check
    success_rate = metrics.get('success_rate', 0)
    if success_rate < bounds.min_success_rate:
        violations.append(f"Success rate {success_rate:.4f} below minimum {bounds.min_success_rate}")

    # Error rate check
    error_rate = metrics.get('error_rate', 0)
    if error_rate > bounds.max_error_rate:
        violations.append(f"Error rate {error_rate:.4f} exceeds maximum {bounds.max_error_rate}")

    # Enhanced metrics validation
    memory_growth = metrics.get('memory_growth_mb', 0)
    if memory_growth > bounds.max_memory_growth_mb:
        violations.append(f"Memory growth {memory_growth:.2f}MB exceeds maximum {bounds.max_memory_growth_mb}MB")

    # Batch efficiency validation (only for consume operations, poll operations don't have batch metrics)
    # Check if this is a consume operation by looking for consume-specific metrics
    if 'messages_per_consume' in metrics:
        messages_per_consume = metrics.get('messages_per_consume', 0)
        if messages_per_consume < bounds.min_messages_per_consume:
            violations.append(f"Messages per consume {messages_per_consume:.2f} "
                              f"below minimum {bounds.min_messages_per_consume}")

        empty_consume_rate = metrics.get('empty_consume_rate', 0)
        if empty_consume_rate > bounds.max_empty_consume_rate:
            violations.append(f"Empty consume rate {empty_consume_rate:.3f} "
                              f"exceeds maximum {bounds.max_empty_consume_rate}")

    # For poll operations, we skip batch efficiency validation since they're single-message operations
    is_valid = len(violations) == 0
    return is_valid, violations


def print_consumer_metrics_report(metrics: Dict[str, Any], is_valid: bool, violations: List[str],
                                  consumer_type: str = None, batch_size: int = None, serialization_type: str = None):
    """Print simplified consumer metrics report"""
    # Detect operation type from metrics keys
    op_name = "consume"  # default
    if any(key.startswith('poll_') for key in metrics.keys()):
        op_name = "poll"

    # Build informative header
    header_parts = ["Consumer Performance Report"]
    if consumer_type:
        header_parts.append(f"{consumer_type.upper()}")
    if op_name == "consume" and batch_size is not None:
        header_parts.append(f"{op_name.upper()}(batch_size={batch_size})")
    if serialization_type:
        header_parts.append(f"serialization_type={serialization_type}")
    else:
        header_parts.append(f"{op_name.upper()}()")

    header = " - ".join(header_parts)
    print(f"\n=== {header} ===")

    # Basic metrics
    print(f"Duration: {metrics.get('duration_seconds', 0):.2f}s")
    print(f"Messages consumed: {metrics.get('messages_consumed', 0)}")
    print(f"Consumption throughput: {metrics.get('consumption_rate_msg_per_sec', 0):.2f} msg/s")
    print(f"Data throughput: {metrics.get('data_throughput_mb_per_sec', 0):.4f} MB/s")

    # Latency metrics
    print("\nLatency Metrics:")
    print(f"  Average: {metrics.get('avg_latency_ms', 0):.2f}ms")
    print(f"  P50: {metrics.get('p50_latency_ms', 0):.2f}ms")
    print(f"  P95: {metrics.get('p95_latency_ms', 0):.2f}ms")
    print(f"  P99: {metrics.get('p99_latency_ms', 0):.2f}ms")
    print(f"  Max: {metrics.get('max_latency_ms', 0):.2f}ms")

    # Reliability metrics
    print("\nReliability:")
    print(f"  Error rate: {metrics.get('error_rate', 0):.4f}")

    # Message size analysis
    print("\nMessage Size Analysis:")
    print(f"  Average: {metrics.get('avg_message_size_bytes', 0):.0f} bytes")
    print(f"  Median: {metrics.get('median_message_size_bytes', 0):.0f} bytes")
    print(f"  Range: {metrics.get('min_message_size_bytes', 0):.0f} - "
          f"{metrics.get('max_message_size_bytes', 0):.0f} bytes")

    # Memory usage (if available)
    if metrics.get('peak_memory_mb', 0) > 0:
        print("\nMemory Usage:")
        print(f"  Peak: {metrics.get('peak_memory_mb', 0):.2f}MB")
        print(f"  Growth: {metrics.get('memory_growth_mb', 0):.2f}MB")

    # Batch efficiency (only for consume operations, not poll)
    # Try to detect operation type from metric keys
    op_name = "consume"  # default
    if any(key.startswith('poll_') for key in metrics.keys()):
        op_name = "poll"

    # Only show efficiency metrics for batch operations (consume), not single-message operations (poll)
    if op_name == "consume":
        print(f"\n{op_name.title()} Efficiency:")
        print(f"  Messages per {op_name}(): {metrics.get(f'messages_per_{op_name}', 0):.2f}")
        print(f"  Empty {op_name} rate: {metrics.get(f'empty_{op_name}_rate', 0):.3f}")
        print(f"  Total {op_name}() calls: {metrics.get(f'{op_name}_call_count', 0)}")

    # Validation
    print("\nValidation:")
    if is_valid:
        print("  All performance bounds satisfied")
    else:
        print("  Performance bounds violations:")
        for violation in violations:
            print(f"    - {violation}")

    print("=" * 50)



# -*- coding: utf-8 -*-
# Copyright 2024 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from confluent_kafka.admin import OffsetSpec, DeletedRecords
from confluent_kafka import TopicPartition


def test_delete_records(kafka_cluster):
    """
    Test delete_records, delete the records upto the specified offset
    in that particular partition of the specified topic.
    """
    admin_client = kafka_cluster.admin()

    # Create a topic with a single partition
    topic = kafka_cluster.create_topic_and_wait_propogation("test-del-records",
                                                            {
                                                                "num_partitions": 1,
                                                                "replication_factor": 1,
                                                            })

    # Create Producer instance
    p = kafka_cluster.producer()
    p.produce(topic, "Message-1")
    p.produce(topic, "Message-2")
    p.produce(topic, "Message-3")
    p.flush()

    topic_partition = TopicPartition(topic, 0)
    requests = {topic_partition: OffsetSpec.earliest()}

    # Check if the earliest avilable offset for this topic partition is 0
    fs = admin_client.list_offsets(requests)
    result = list(fs.values())[0].result()
    assert (result.offset == 0)

    topic_partition_offset = TopicPartition(topic, 0, 2)

    # Delete the records
    fs1 = admin_client.delete_records([topic_partition_offset])

    # Find the earliest available offset for that specific topic partition after deletion has been done
    fs2 = admin_client.list_offsets(requests)

    # Check if the earliest available offset is equal to the offset passed to the delete records function
    res = list(fs1.values())[0].result()
    assert isinstance(res, DeletedRecords)
    assert (res.low_watermark == list(fs2.values())[0].result().offset)

    # Delete created topic
    fs = admin_client.delete_topics([topic])
    for topic, f in fs.items():
        f.result()


def test_delete_records_multiple_topics_and_partitions(kafka_cluster):
    """
    Test delete_records, delete the records upto the specified offset
    in that particular partition of the specified topic.
    """
    admin_client = kafka_cluster.admin()
    num_partitions = 3
    # Create two topics with a single partition
    topic = kafka_cluster.create_topic_and_wait_propogation("test-del-records",
                                                            {
                                                                "num_partitions": num_partitions,
                                                                "replication_factor": 1,
                                                            })
    topic2 = kafka_cluster.create_topic_and_wait_propogation("test-del-records2",
                                                             {
                                                                 "num_partitions": num_partitions,
                                                                 "replication_factor": 1,
                                                             })

    topics = [topic, topic2]
    partitions = list(range(num_partitions))
    # Create Producer instance
    p = kafka_cluster.producer()
    for t in topics:
        for partition in partitions:
            p.produce(t, "Message-1", partition=partition)
            p.produce(t, "Message-2", partition=partition)
            p.produce(t, "Message-3", partition=partition)
    p.flush()
    requests = dict(
        [
            (TopicPartition(t, partition), OffsetSpec.earliest())
            for t in topics
            for partition in partitions
        ]
    )
    # Check if the earliest available offset for this topic partition is 0
    fs = admin_client.list_offsets(requests)
    assert all([p.result().offset == 0 for p in fs.values()])
    delete_index = 0
    # Delete the records
    for delete_partitions in [
        # Single partition no deletion
        [TopicPartition(topic, 0, 0)],
        # Single topic, two partitions, single record deleted
        [TopicPartition(topic, 0, 1), TopicPartition(topic, 1, 1)],
        # Two topics, four partitions, two records deleted
        [TopicPartition(topic, 2, 2), TopicPartition(topic2, 0, 2),
         TopicPartition(topic2, 1, 2), TopicPartition(topic2, 2, 2)],
    ]:
        list_offsets_requests = dict([
            (part, OffsetSpec.earliest()) for part in delete_partitions
        ])
        futmap_delete = admin_client.delete_records(delete_partitions)
        delete_results = [(part, fut.result())
                          for part, fut in futmap_delete.items()]
        futmap_list = admin_client.list_offsets(list_offsets_requests)
        list_results = dict([(part, fut.result())
                            for part, fut in futmap_list.items()])
        for part, delete_result in delete_results:
            list_result = list_results[part]
            assert isinstance(delete_result, DeletedRecords)
            assert delete_result.low_watermark == list_result.offset
            assert delete_result.low_watermark == delete_index
        delete_index += 1

    # Delete created topics
    fs = admin_client.delete_topics(topics)
    for topic, f in fs.items():
        f.result()



"""
Producer strategies for testing sync and async Kafka producers.

This module contains strategy classes that encapsulate the different producer
implementations (sync vs async) with consistent interfaces for testing.
"""
import time
import asyncio
import uuid
from confluent_kafka import Producer
import json
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry._sync.json_schema import JSONSerializer
from confluent_kafka.schema_registry._sync.protobuf import ProtobufSerializer
from confluent_kafka.schema_registry.avro import AvroSerializer
from confluent_kafka.schema_registry._async.json_schema import AsyncJSONSerializer
from confluent_kafka.schema_registry._async.protobuf import AsyncProtobufSerializer
from confluent_kafka.schema_registry._async.avro import AsyncAvroSerializer
from confluent_kafka.schema_registry.schema_registry_client import AsyncSchemaRegistryClient
from confluent_kafka.serialization import MessageField, SerializationContext, StringSerializer
from tests.integration.schema_registry.data.proto import PublicTestProto_pb2
import os


class ProducerStrategy:
    """Base class for producer strategies"""

    def __init__(self, bootstrap_servers, logger):
        self.bootstrap_servers = bootstrap_servers
        self.logger = logger
        self.metrics = None

    def create_producer(self):
        raise NotImplementedError()

    def produce_messages(self, topic_name, test_duration, start_time, message_formatter,
                         delivered_container, failed_container=None, serialization_type=None):
        raise NotImplementedError()

    def _get_base_config(self):
        """Get shared Kafka producer configuration optimized for low-latency, high-throughput"""
        return {
            'bootstrap.servers': self.bootstrap_servers,
            'queue.buffering.max.messages': 1000000,  # 1M messages (sufficient)
            'queue.buffering.max.kbytes': 1048576,    # 1GB (default)
            'batch.size': 65536,                      # 64KB batches (increased for better efficiency)
            'batch.num.messages': 50000,              # 50K messages per batch (up from 10K default)
            'message.max.bytes': 2097152,             # 2MB max message size (up from ~1MB default)
            'linger.ms': 1,                          # Wait 1ms for batching (low latency)
            'compression.type': 'lz4',               # Fast compression
            'acks': 1,                               # Wait for leader only (faster)
            'retries': 3,                            # Retry failed sends
            'delivery.timeout.ms': 30000,            # 30s delivery timeout
            'max.in.flight.requests.per.connection': 5  # Pipeline requests
        }

    def _log_configuration(self, config, producer_type, extra_params=None):
        """Log producer configuration for validation"""
        separator = "=" * (len(f"{producer_type.upper()} PRODUCER CONFIGURATION") + 6)
        self.logger.info(f"=== {producer_type.upper()} PRODUCER CONFIGURATION ===")
        for key, value in config.items():
            self.logger.info(f"{key}: {value}")

        if extra_params:
            for key, value in extra_params.items():
                self.logger.info(f"{key}: {value}")

        self.logger.info(separator)

    def _print_timing_metrics(self, producer_type, produce_times, poll_times, flush_time):
        """Print code path timing metrics"""
        avg_produce_time = sum(produce_times) / len(produce_times) * 1000 if produce_times else 0
        avg_poll_time = sum(poll_times) / len(poll_times) * 1000 if poll_times else 0
        flush_time_ms = flush_time * 1000

        separator = "=" * (len(f"{producer_type.upper()} PRODUCER CODE PATH TIMING") + 6)
        print(f"\n=== {producer_type.upper()} PRODUCER CODE PATH TIMING ===")
        print(f"Time to call {'AIO' if producer_type == 'ASYNC' else ''}Producer.produce(): {avg_produce_time:.4f}ms")
        print(f"Time to call {'AIO' if producer_type == 'ASYNC' else ''}Producer.poll(): {avg_poll_time:.4f}ms")
        print(f"Time to call {'AIO' if producer_type == 'ASYNC' else ''}Producer.flush(): {flush_time_ms:.4f}ms")
        print(f"Total produce() calls: {len(produce_times)}")
        print(f"Total poll() calls: {len(poll_times)}")
        print(separator)

    def build_serializers(self, serialization_type):
        """Build and return (key_serializer, value_serializer) for the given serialization_type.
        Returns (None, None) if serialization_type is not provided.
        This method should be implemented by concrete strategy classes.
        """
        raise NotImplementedError("Subclasses must implement build_serializers")


class SyncProducerStrategy(ProducerStrategy):
    def build_serializers(self, serialization_type):
        """Build synchronous serializers for Schema Registry"""
        if not serialization_type:
            return None, None

        schema_registry_url = os.getenv(
            'SCHEMA_REGISTRY_URL',
            getattr(self, 'schema_registry_url', 'http://localhost:8081')
        )
        client_config = {
            'url': schema_registry_url,
        }
        sr_client = SchemaRegistryClient(client_config)
        key_serializer = StringSerializer('utf8')

        if serialization_type == 'avro':
            avro_schema = {
                "type": "record",
                "name": "TestMessage",
                "fields": [
                    {"name": "test_string", "type": "string"},
                    {"name": "test_bool", "type": "boolean"},
                    {"name": "test_bytes", "type": "bytes"},
                    {"name": "test_double", "type": "double"},
                    {"name": "test_float", "type": "float"},
                    {"name": "test_fixed32", "type": "int"},
                    {"name": "test_fixed64", "type": "long"},
                    {"name": "test_int32", "type": "int"},
                    {"name": "test_int64", "type": "long"},
                    {"name": "test_sfixed32", "type": "int"},
                    {"name": "test_sfixed64", "type": "long"},
                    {"name": "test_sint32", "type": "int"},
                    {"name": "test_sint64", "type": "long"},
                    {"name": "test_uint32", "type": "int"},
                    {"name": "test_uint64", "type": "long"}
                ]
            }
            value_serializer = AvroSerializer(
                schema_registry_client=sr_client,
                schema_str=json.dumps(avro_schema)
            )
        elif serialization_type == 'json':
            json_schema = {
                "type": "object",
                "title": "TestMessage",  # Required for Schema Registry
                "properties": {
                    "test_string": {"type": "string"},
                    "test_bool": {"type": "boolean"},
                    "test_bytes": {"type": "string"},  # JSON represents bytes as string
                    "test_double": {"type": "number"},
                    "test_float": {"type": "number"},
                    "test_fixed32": {"type": "integer"},
                    "test_fixed64": {"type": "integer"},
                    "test_int32": {"type": "integer"},
                    "test_int64": {"type": "integer"},
                    "test_sfixed32": {"type": "integer"},
                    "test_sfixed64": {"type": "integer"},
                    "test_sint32": {"type": "integer"},
                    "test_sint64": {"type": "integer"},
                    "test_uint32": {"type": "integer"},
                    "test_uint64": {"type": "integer"}
                },
                "required": [
                    "test_string", "test_bool", "test_bytes", "test_double", "test_float",
                    "test_fixed32", "test_fixed64", "test_int32", "test_int64",
                    "test_sfixed32", "test_sfixed64", "test_sint32", "test_sint64",
                    "test_uint32", "test_uint64"
                ]
            }
            value_serializer = JSONSerializer(json.dumps(json_schema), sr_client)
        elif serialization_type == 'protobuf':
            value_serializer = ProtobufSerializer(PublicTestProto_pb2.TestMessage, sr_client)
        else:
            raise ValueError(f"Unsupported serialization_type: {serialization_type}")

        return key_serializer, value_serializer

    def create_producer(self, config_overrides=None):
        config = self._get_base_config()

        # Apply any test-specific overrides
        if config_overrides:
            config.update(config_overrides)

        producer = Producer(config)

        # Log the configuration for validation
        self._log_configuration(config, "SYNC")

        return producer

    def create_transactional_producer(self):
        overrides = {
            'transactional.id': f'sync-tx-producer-{uuid.uuid4()}',
            'acks': 'all',
            'enable.idempotence': True
        }
        return self.create_producer(config_overrides=overrides)

    def produce_messages(self, topic_name, test_duration, start_time, message_formatter,
                         delivered_container, failed_container=None, serialization_type=None,
                         use_transaction=False):
        config_overrides = getattr(self, 'config_overrides', None)
        if use_transaction:
            producer = self.create_transactional_producer()
            producer.init_transactions()
        else:
            producer = self.create_producer(config_overrides)

        # Temporary metrics for timing sections
        messages_sent = 0
        send_times = {}  # Track send times for latency calculation
        produce_times = []
        poll_times = []
        flush_time = 0

        # Initialize serializers via helper (or None if not requested)
        key_serializer, value_serializer = self.build_serializers(serialization_type)

        def delivery_callback(err, msg):
            if err:
                if failed_container is not None:
                    failed_container.append(err)
                if self.metrics:
                    self.metrics.record_failed(
                        topic=msg.topic() if msg else topic_name,
                        partition=msg.partition() if msg else 0
                    )
            else:
                delivered_container.append(msg)
                if self.metrics:
                    # Calculate latency if we have send time
                    msg_key = msg.key().decode('utf-8', errors='replace') if msg.key() else 'unknown'
                    latency_ms = 0.0
                    if msg_key in send_times:
                        latency_ms = (time.time() - send_times[msg_key]) * 1000
                        del send_times[msg_key]

                    self.metrics.record_delivered(latency_ms, topic=msg.topic(), partition=msg.partition())

        while time.time() - start_time < test_duration:
            message_value, message_key = message_formatter(messages_sent)
            try:
                # Begin transaction if using transactions
                if use_transaction:
                    producer.begin_transaction()

                if serialization_type:
                    # Serialize key and value using Schema Registry serializers
                    serialized_key = key_serializer(message_key)
                    serialized_value = value_serializer(
                        message_value, SerializationContext(topic_name, MessageField.VALUE)
                    )
                    message_size = len(serialized_key) + len(serialized_value)
                    if self.metrics:
                        send_times[message_key] = time.time()
                        self.metrics.record_sent(message_size, topic=topic_name, partition=0)
                    produce_start = time.time()
                    producer.produce(
                        topic=topic_name,
                        value=serialized_value,
                        key=serialized_key,
                        on_delivery=delivery_callback
                    )
                else:
                    # Track send time for latency calculation (plain string messages)
                    if self.metrics:
                        send_times[message_key] = time.time()
                        message_size = len(message_value.encode('utf-8')) + len(message_key.encode('utf-8'))
                        self.metrics.record_sent(message_size, topic=topic_name, partition=0)
                    # Produce message
                    produce_start = time.time()
                    producer.produce(
                        topic=topic_name,
                        value=message_value,
                        key=message_key,
                        on_delivery=delivery_callback
                    )
                # Commit transaction if using transactions
                if use_transaction:
                    producer.commit_transaction()

                produce_times.append(time.time() - produce_start)
                messages_sent += 1

                # Use configured polling interval (default to 50 if not set)
                poll_interval = getattr(self, 'poll_interval', 50)

                if messages_sent % poll_interval == 0:
                    poll_start = time.time()
                    producer.poll(0)
                    poll_times.append(time.time() - poll_start)
                    if self.metrics:
                        self.metrics.record_poll()

            except Exception as e:
                if failed_container is not None:
                    failed_container.append(e)
                if self.metrics:
                    self.metrics.record_failed(topic=topic_name, partition=0)
                self.logger.error(f"Failed to produce message {messages_sent}: {e}")

        # Flush producer
        flush_start = time.time()
        producer.flush(timeout=30)
        flush_time = time.time() - flush_start

        # Print timing metrics
        self._print_timing_metrics("SYNC", produce_times, poll_times, flush_time)

        return messages_sent

    def get_final_metrics(self):
        """Return final metrics summary for the sync producer"""
        if self.metrics:
            return self.metrics
        return None


class AsyncProducerStrategy(ProducerStrategy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._producer_instance = None

    async def build_serializers(self, serialization_type):
        """Build asynchronous serializers for Schema Registry"""
        if not serialization_type:
            return None, None

        schema_registry_url = os.getenv(
            'SCHEMA_REGISTRY_URL',
            getattr(self, 'schema_registry_url', 'http://localhost:8081')
        )
        client_config = {
            'url': schema_registry_url,
        }
        sr_client = AsyncSchemaRegistryClient(client_config)
        key_serializer = StringSerializer('utf8')

        if serialization_type == 'avro':
            avro_schema = {
                "type": "record",
                "name": "TestMessage",
                "fields": [
                    {"name": "test_string", "type": "string"},
                    {"name": "test_bool", "type": "boolean"},
                    {"name": "test_bytes", "type": "bytes"},
                    {"name": "test_double", "type": "double"},
                    {"name": "test_float", "type": "float"},
                    {"name": "test_fixed32", "type": "int"},
                    {"name": "test_fixed64", "type": "long"},
                    {"name": "test_int32", "type": "int"},
                    {"name": "test_int64", "type": "long"},
                    {"name": "test_sfixed32", "type": "int"},
                    {"name": "test_sfixed64", "type": "long"},
                    {"name": "test_sint32", "type": "int"},
                    {"name": "test_sint64", "type": "long"},
                    {"name": "test_uint32", "type": "int"},
                    {"name": "test_uint64", "type": "long"}
                ]
            }
            value_serializer = await AsyncAvroSerializer(
                schema_registry_client=sr_client,
                schema_str=json.dumps(avro_schema)
            )
        elif serialization_type == 'json':
            json_schema = {
                "type": "object",
                "title": "TestMessage",  # Required for Schema Registry
                "properties": {
                    "test_string": {"type": "string"},
                    "test_bool": {"type": "boolean"},
                    "test_bytes": {"type": "string"},
                    "test_double": {"type": "number"},
                    "test_float": {"type": "number"},
                    "test_fixed32": {"type": "integer"},
                    "test_fixed64": {"type": "integer"},
                    "test_int32": {"type": "integer"},
                    "test_int64": {"type": "integer"},
                    "test_sfixed32": {"type": "integer"},
                    "test_sfixed64": {"type": "integer"},
                    "test_sint32": {"type": "integer"},
                    "test_sint64": {"type": "integer"},
                    "test_uint32": {"type": "integer"},
                    "test_uint64": {"type": "integer"}
                },
                "required": [
                    "test_string", "test_bool", "test_bytes", "test_double", "test_float",
                    "test_fixed32", "test_fixed64", "test_int32", "test_int64",
                    "test_sfixed32", "test_sfixed64", "test_sint32", "test_sint64",
                    "test_uint32", "test_uint64"
                ]
            }
            value_serializer = await AsyncJSONSerializer(json.dumps(json_schema), sr_client)
        elif serialization_type == 'protobuf':
            value_serializer = await AsyncProtobufSerializer(PublicTestProto_pb2.TestMessage, sr_client)
        else:
            raise ValueError(f"Unsupported serialization_type: {serialization_type}")

        return key_serializer, value_serializer

    def create_producer(self, config_overrides=None):
        from confluent_kafka.experimental.aio import AIOProducer
        # Enable logging for AIOProducer
        import logging
        logging.basicConfig(level=logging.INFO)

        config = self._get_base_config()

        # Apply any test-specific overrides
        if config_overrides:
            config.update(config_overrides)

        # Get producer configuration from strategy attributes
        max_workers = getattr(self, 'max_workers', 4)
        batch_size = getattr(self, 'batch_size', 1000)  # Optimal batch size for low latency

        # Use updated defaults with configurable parameters
        self._producer_instance = AIOProducer(config, max_workers=max_workers, batch_size=batch_size)

        # Log the configuration for validation
        extra_params = {'max_workers': max_workers, 'batch_size': batch_size}
        self._log_configuration(config, "ASYNC", extra_params)

        return self._producer_instance

    def create_transactional_producer(self):
        overrides = {
            'transactional.id': f'async-tx-producer-{uuid.uuid4()}',
            'acks': 'all',
            'enable.idempotence': True
        }
        return self.create_producer(config_overrides=overrides)

    def produce_messages(self, topic_name, test_duration, start_time, message_formatter,
                         delivered_container, failed_container=None, serialization_type=None,
                         use_transaction=False):

        async def async_produce():
            config_overrides = getattr(self, 'config_overrides', None)

            if use_transaction:
                producer = self.create_transactional_producer()
                await producer.init_transactions()
            else:
                producer = self.create_producer(config_overrides)

            # Temporary metrics for timing sections
            messages_sent = 0
            produce_times = []
            poll_times = []  # Async producer doesn't use polling, but needed for metrics
            flush_time = 0
            pending_futures = []
            send_times = {}

            # Get serializers if using Schema Registry
            if serialization_type:
                key_serializer, value_serializer = await self.build_serializers(serialization_type)

            # Pre-create shared metrics callback to avoid closure creation overhead
            if self.metrics:
                def shared_metrics_callback(err, msg):
                    if not err:
                        # Calculate latency if we have send time
                        msg_key = msg.key().decode('utf-8', errors='replace') if msg.key() else 'unknown'
                        latency_ms = 0.0
                        if msg_key in send_times:
                            latency_ms = (time.time() - send_times[msg_key]) * 1000
                            del send_times[msg_key]
                        self.metrics.record_delivered(latency_ms, topic=msg.topic(), partition=msg.partition())
                    if not err:  # Also append to delivered_messages for assertion
                        delivered_container.append(msg)

            while time.time() - start_time < test_duration:
                message_value, message_key = message_formatter(messages_sent)
                try:
                    # Begin transaction if using transactions
                    if use_transaction:
                        await producer.begin_transaction()

                    # Handle serialization if using Schema Registry
                    if serialization_type:
                        serialized_key = key_serializer(message_key)
                        # Async serializers return coroutines, so we need to await them
                        serialized_value = await value_serializer(
                            message_value, SerializationContext(topic_name, MessageField.VALUE)
                        )
                        message_size = len(serialized_key) + len(serialized_value)
                        produce_value = serialized_value
                        produce_key = serialized_key
                    else:
                        # Plain string encoding for non-SR
                        message_size = len(message_value.encode('utf-8')) + len(message_key.encode('utf-8'))
                        produce_value = message_value
                        produce_key = message_key

                    # Record sent message for metrics and track send time
                    if self.metrics:
                        self.metrics.record_sent(message_size, topic=topic_name, partition=0)
                        send_times[message_key] = time.time()  # Track send time for latency

                    # Produce message
                    produce_start = time.time()
                    delivery_future = await producer.produce(
                        topic=topic_name,
                        value=produce_value,
                        key=produce_key,
                        on_delivery=shared_metrics_callback
                    )

                    # Commit transaction if using transactions
                    if use_transaction:
                        await producer.commit_transaction()

                    produce_times.append(time.time() - produce_start)
                    pending_futures.append((delivery_future, message_key))  # Store delivery future
                    messages_sent += 1

                    # Use configured polling interval (default to 50 if not set)
                    poll_interval = getattr(self, 'poll_interval', 50)

                    if messages_sent % poll_interval == 0:
                        poll_start = time.time()
                        await producer.poll(0)
                        poll_times.append(time.time() - poll_start)
                        if self.metrics:
                            self.metrics.record_poll()

                except Exception as e:
                    if failed_container is not None:
                        failed_container.append(e)
                    if self.metrics:
                        self.metrics.record_failed(topic=topic_name, partition=0)
                    self.logger.error(f"Failed to produce message {messages_sent}: {e}")

            # Flush producer
            flush_start = time.time()
            await producer.flush(timeout=30)
            flush_time = time.time() - flush_start

            # Wait for all pending futures to complete (for delivery confirmation only)
            for delivery_future, message_key in pending_futures:
                try:
                    msg = await delivery_future
                    delivered_container.append(msg)

                    # Record delivery metrics (replaces the old callback approach)
                    if self.metrics:
                        # Calculate latency if we have send time
                        latency_ms = 0.0
                        if message_key in send_times:
                            latency_ms = (time.time() - send_times[message_key]) * 1000
                            del send_times[message_key]

                        self.metrics.record_delivered(latency_ms, topic=msg.topic(), partition=msg.partition())

                except Exception as e:
                    if failed_container is not None:
                        failed_container.append(e)
                    if self.metrics:
                        self.metrics.record_failed(topic=topic_name, partition=0)
                    self.logger.error(f"Failed to deliver message with key {message_key}: {e}")

            # Print timing metrics
            self._print_timing_metrics("ASYNC", produce_times, poll_times, flush_time)

            # Close producer to ensure clean shutdown
            await producer.close()

            return messages_sent

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(async_produce())

    def get_final_metrics(self):
        """Return final metrics summary for the async producer"""
        if self.metrics:
            return self.metrics
        return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tests/integration/schema_registry/data/proto/DependencyTestProto.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tests.integration.schema_registry.data.proto import NestedTestProto_pb2 as tests_dot_integration_dot_schema__registry_dot_data_dot_proto_dot_NestedTestProto__pb2
from tests.integration.schema_registry.data.proto import PublicTestProto_pb2 as tests_dot_integration_dot_schema__registry_dot_data_dot_proto_dot_PublicTestProto__pb2
try:
  tests_dot_integration_dot_schema__registry_dot_data_dot_proto_dot_TestProto__pb2 = tests_dot_integration_dot_schema__registry_dot_data_dot_proto_dot_PublicTestProto__pb2.tests_dot_integration_dot_schema__registry_dot_data_dot_proto_dot_TestProto__pb2
except AttributeError:
  tests_dot_integration_dot_schema__registry_dot_data_dot_proto_dot_TestProto__pb2 = tests_dot_integration_dot_schema__registry_dot_data_dot_proto_dot_PublicTestProto__pb2.tests.integration.schema_registry.data.proto.TestProto_pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nFtests/integration/schema_registry/data/proto/DependencyTestProto.proto\x12$tests.integration.serialization.data\x1a\x42tests/integration/schema_registry/data/proto/NestedTestProto.proto\x1a\x42tests/integration/schema_registry/data/proto/PublicTestProto.proto\"\x98\x01\n\x11\x44\x65pendencyMessage\x12K\n\x0enested_message\x18\x01 \x01(\x0b\x32\x33.tests.integration.serialization.data.NestedMessage\x12\x11\n\tis_active\x18\x02 \x01(\x08\x12#\n\rtest_messsage\x18\x03 \x01(\x0b\x32\x0c.TestMessageB.\n,io.confluent.kafka.serializers.protobuf.testb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tests.integration.schema_registry.data.proto.DependencyTestProto_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n,io.confluent.kafka.serializers.protobuf.test'
  _DEPENDENCYMESSAGE._serialized_start=249
  _DEPENDENCYMESSAGE._serialized_end=401
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2024 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


def test_wildcard_matcher():
    from confluent_kafka.schema_registry.wildcard_matcher import wildcard_match

    assert not wildcard_match("", "Foo")
    assert not wildcard_match("Foo", "")
    assert wildcard_match("", "")
    assert wildcard_match("Foo", "Foo")
    assert wildcard_match("", "*")
    assert not wildcard_match("", "?")
    assert wildcard_match("Foo", "Fo*")
    assert wildcard_match("Foo", "Fo?")
    assert wildcard_match("Foo Bar and Catflag", "Fo*")
    assert wildcard_match("New Bookmarks", "N?w ?o?k??r?s")
    assert not wildcard_match("Foo", "Bar")
    assert wildcard_match("Foo Bar Foo", "F*o Bar*")
    assert wildcard_match("Adobe Acrobat Installer", "Ad*er")
    assert wildcard_match("Foo", "*Foo")
    assert wildcard_match("BarFoo", "*Foo")
    assert wildcard_match("Foo", "Foo*")
    assert wildcard_match("FooBar", "Foo*")
    assert not wildcard_match("FOO", "*Foo")
    assert not wildcard_match("BARFOO", "*Foo")
    assert not wildcard_match("FOO", "Foo*")
    assert not wildcard_match("FOOBAR", "Foo*")
    assert wildcard_match("eve", "eve*")
    assert wildcard_match("alice.bob.eve", "a*.bob.eve")
    assert not wildcard_match("alice.bob.eve", "a*")
    assert wildcard_match("alice.bob.eve", "a**")
    assert not wildcard_match("alice.bob.eve", "alice.bob*")
    assert wildcard_match("alice.bob.eve", "alice.bob**")



# -*- coding: utf-8 -*-
# Copyright 2023 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import concurrent
import pytest

from confluent_kafka.admin import UserScramCredentialsDescription, UserScramCredentialUpsertion, \
    UserScramCredentialDeletion, ScramCredentialInfo, \
    ScramMechanism
from confluent_kafka.error import KafkaException, KafkaError

from tests.common import TestUtils


def test_user_scram_credentials(kafka_cluster):
    """
    Tests for the alter and describe SASL/SCRAM credential operations.
    """

    admin_client = kafka_cluster.admin()

    newuser = "non-existent"
    mechanism = ScramMechanism.SCRAM_SHA_256
    iterations = 10000
    password = b"password"
    salt = b"salt"

    futmap = admin_client.describe_user_scram_credentials([newuser])
    assert isinstance(futmap, dict)
    assert len(futmap) == 1
    assert newuser in futmap
    fut = futmap[newuser]
    with pytest.raises(KafkaException) as ex:
        result = fut.result()
    assert ex.value.args[0] == KafkaError.RESOURCE_NOT_FOUND

    # Insert new user with SCRAM_SHA_256: 10000
    futmap = admin_client.alter_user_scram_credentials([UserScramCredentialUpsertion(newuser,
                                                        ScramCredentialInfo(mechanism, iterations),
                                                        password, salt)])
    fut = futmap[newuser]
    result = fut.result()
    assert result is None

    time.sleep(1)

    # Try upsertion for newuser,SCRAM_SHA_256 and add newuser,SCRAM_SHA_512
    request = [UserScramCredentialUpsertion(newuser,
                                            ScramCredentialInfo(
                                                mechanism, iterations),
                                            password, salt),
               UserScramCredentialUpsertion(newuser,
                                            ScramCredentialInfo(
                                                ScramMechanism.SCRAM_SHA_512, 10000),
                                            password)]

    if TestUtils.use_kraft():
        futmap = admin_client.alter_user_scram_credentials([request[0]])
        result = futmap[newuser].result()
        assert result is None
        futmap = admin_client.alter_user_scram_credentials([request[1]])
    else:
        futmap = admin_client.alter_user_scram_credentials(request)
    fut = futmap[newuser]
    result = fut.result()
    assert result is None

    time.sleep(1)

    # Delete newuser,SCRAM_SHA_512
    futmap = admin_client.alter_user_scram_credentials([UserScramCredentialDeletion(
        newuser,
        ScramMechanism.SCRAM_SHA_512)
    ])
    fut = futmap[newuser]
    result = fut.result()
    assert result is None

    time.sleep(1)

    # Describe all users
    for args in [[], [None]]:
        f = admin_client.describe_user_scram_credentials(*args)
        assert isinstance(f, concurrent.futures.Future)
        results = f.result()
        assert newuser in results
        description = results[newuser]
        assert isinstance(description, UserScramCredentialsDescription)
        for scram_credential_info in description.scram_credential_infos:
            assert ((scram_credential_info.mechanism == mechanism) and
                    (scram_credential_info.iterations == iterations))

    # Describe specific user
    futmap = admin_client.describe_user_scram_credentials([newuser])
    assert isinstance(futmap, dict)
    assert len(futmap) == 1
    assert newuser in futmap
    description = futmap[newuser].result()
    assert isinstance(description, UserScramCredentialsDescription)
    for scram_credential_info in description.scram_credential_infos:
        assert ((scram_credential_info.mechanism == mechanism) and
                (scram_credential_info.iterations == iterations))

    # Delete newuser
    futmap = admin_client.alter_user_scram_credentials([UserScramCredentialDeletion(newuser, mechanism)])
    assert isinstance(futmap, dict)
    assert len(futmap) == 1
    assert newuser in futmap
    fut = futmap[newuser]
    result = fut.result()
    assert result is None

    time.sleep(1)

    # newuser isn't found
    futmap = admin_client.describe_user_scram_credentials([newuser])
    assert isinstance(futmap, dict)
    assert len(futmap) == 1
    assert newuser in futmap
    fut = futmap[newuser]
    with pytest.raises(KafkaException) as ex:
        result = fut.result()
    assert ex.value.args[0] == KafkaError.RESOURCE_NOT_FOUND



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2025 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Examples of setting up Schema Registry with OAuth with static token,
# Client Credentials, and custom functions


# CUSTOM OAuth configuration takes in a custom function, config for that
# function, and returns the fields shown below. All fields must be returned
# for OAuth authentication to work


from confluent_kafka.schema_registry.schema_registry_client import SchemaRegistryClient


def main():
    static_oauth_config = {'url': 'https://psrc-123456.us-east-1.aws.confluent.cloud',
                           'bearer.auth.credentials.source': 'STATIC_TOKEN',
                           'bearer.auth.token': 'static-token',
                           'bearer.auth.logical.cluster': 'lsrc-12345',
                           'bearer.auth.identity.pool.id': 'pool-abcd'}
    static_oauth_sr_client = SchemaRegistryClient(static_oauth_config)
    print(static_oauth_sr_client.get_subjects())

    client_credentials_oauth_config = {
        'url': 'https://psrc-123456.us-east-1.aws.confluent.cloud',
        'bearer.auth.credentials.source': 'OAUTHBEARER',
        'bearer.auth.client.id': 'client-id',
        'bearer.auth.client.secret': 'client-secret',
        'bearer.auth.scope': 'schema_registry',
        'bearer.auth.issuer.endpoint.url': 'https://yourauthprovider.com/v1/token',
        'bearer.auth.logical.cluster': 'lsrc-12345',
        'bearer.auth.identity.pool.id': 'pool-abcd'}

    client_credentials_oauth_sr_client = SchemaRegistryClient(client_credentials_oauth_config)
    print(client_credentials_oauth_sr_client.get_subjects())

    def custom_oauth_function(config):
        return config

    custom_config = {'bearer.auth.token': 'example-token',
                     'bearer.auth.logical.cluster': 'lsrc-12345', 'bearer.auth.identity.pool.id': 'pool-abcd'}

    custom_sr_config = {'url': 'https://psrc-123456.us-east-1.aws.confluent.cloud',
                        'bearer.auth.credentials.source': 'CUSTOM',
                        'bearer.auth.custom.provider.function': custom_oauth_function,
                        'bearer.auth.custom.provider.config': custom_config}

    custom_sr_client = SchemaRegistryClient(custom_sr_config)
    print(custom_sr_client.get_subjects())


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Unit tests for the BatchProcessor class (_batch_processor.py)

This module tests the BatchProcessor class to ensure proper
message batching, topic grouping, and future management.
"""

from confluent_kafka.experimental.aio.producer._kafka_batch_executor import ProducerBatchExecutor as KafkaBatchExecutor
from confluent_kafka.experimental.aio.producer._AIOProducer import AIOProducer
from confluent_kafka.experimental.aio.producer._producer_batch_processor import (
    ProducerBatchManager as ProducerBatchProcessor
)
import asyncio
import unittest
from unittest.mock import Mock, patch
import sys
import os
import concurrent.futures
import confluent_kafka


# Add src to path for imports
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'src'))


class TestProducerBatchProcessor(unittest.TestCase):
    """Test cases for ProducerBatchProcessor class"""

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)

        self.producer_config = {
            'bootstrap.servers': 'localhost:9092',
            'client.id': 'test-producer',
            'message.timeout.ms': 100,
            'queue.buffering.max.messages': 1000,
            'api.version.request': False,
        }

        self.confluent_kafka_producer = confluent_kafka.Producer(self.producer_config)

        self.kafka_executor = KafkaBatchExecutor(self.confluent_kafka_producer, self.executor)
        self.batch_processor = ProducerBatchProcessor(self.kafka_executor)

        async def create_aio_producer():
            return AIOProducer(self.producer_config, executor=self.executor)

        self.aio_producer = self.loop.run_until_complete(create_aio_producer())

    def tearDown(self):
        try:
            self.loop.run_until_complete(self.aio_producer.close())
        except Exception:
            pass

        try:
            self.confluent_kafka_producer.flush(timeout=1)
        except Exception:
            pass

        try:
            self.executor.shutdown(wait=True, timeout=1)
        except Exception:
            pass

        self.loop.close()

    def test_basic_functionality(self):
        self.assertEqual(self.batch_processor.get_buffer_size(), 0)
        self.assertTrue(self.batch_processor.is_buffer_empty())

        future1 = Mock()
        future2 = Mock()
        msg_data1 = {'topic': 'topic1', 'value': 'test1', 'key': 'key1'}
        msg_data2 = {'topic': 'topic2', 'value': 'test2', 'key': 'key2'}

        self.batch_processor.add_message(msg_data1, future1)
        self.batch_processor.add_message(msg_data2, future2)

        self.assertEqual(self.batch_processor.get_buffer_size(), 2)
        self.assertFalse(self.batch_processor.is_buffer_empty())

        self.batch_processor.clear_buffer()

        self.assertEqual(self.batch_processor.get_buffer_size(), 0)
        self.assertTrue(self.batch_processor.is_buffer_empty())

    def test_group_messages_by_topic(self):
        future1 = Mock()
        future2 = Mock()
        future3 = Mock()

        msg1 = {'topic': 'topic1', 'value': 'test1', 'user_callback': Mock()}
        msg2 = {'topic': 'topic2', 'value': 'test2'}
        msg3 = {'topic': 'topic1', 'value': 'test3', 'user_callback': Mock()}

        self.batch_processor.add_message(msg1, future1)
        self.batch_processor.add_message(msg2, future2)
        self.batch_processor.add_message(msg3, future3)

        topic_groups = self.batch_processor._group_messages_by_topic_and_partition()

        self.assertEqual(len(topic_groups), 2)

        self.assertIn(('topic1', -1), topic_groups)
        topic1_group = topic_groups[('topic1', -1)]
        self.assertEqual(len(topic1_group['messages']), 2)
        self.assertEqual(len(topic1_group['futures']), 2)
        self.assertEqual(topic1_group['futures'][0], future1)
        self.assertEqual(topic1_group['futures'][1], future3)

        self.assertIn(('topic2', -1), topic_groups)
        topic2_group = topic_groups[('topic2', -1)]
        self.assertEqual(len(topic2_group['messages']), 1)
        self.assertEqual(len(topic2_group['futures']), 1)
        self.assertEqual(topic2_group['futures'][0], future2)

    def test_prepare_batch_messages(self):
        messages = [
            {'topic': 'test', 'value': 'test1', 'user_callback': Mock(), 'key': 'key1'},
            {'topic': 'test', 'value': 'test2', 'partition': 1},
        ]

        batch_messages = self.batch_processor._prepare_batch_messages(messages)

        self.assertEqual(len(batch_messages), 2)

        self.assertNotIn('topic', batch_messages[0])
        self.assertIn('value', batch_messages[0])
        self.assertIn('key', batch_messages[0])

        self.assertNotIn('topic', batch_messages[1])
        self.assertIn('value', batch_messages[1])
        self.assertIn('partition', batch_messages[1])

    def test_assign_future_callbacks(self):
        batch_messages = [
            {'value': 'test1'},
            {'value': 'test2'},
        ]
        futures = [Mock(), Mock()]

        self.batch_processor._assign_future_callbacks(batch_messages, futures)

        self.assertIn('callback', batch_messages[0])
        self.assertIn('callback', batch_messages[1])

    def test_handle_batch_failure(self):
        """Test handling batch failures"""
        futures = [Mock(), Mock()]
        futures[0].done.return_value = False
        futures[1].done.return_value = True  # Already done

        exception = RuntimeError("Batch failed")

        # Handle batch failure
        self.batch_processor._handle_batch_failure(
            exception, futures
        )

        # Verify first future got exception (not already done)
        futures[0].set_exception.assert_called_once_with(exception)

        # Verify second future was not modified (already done)
        futures[1].set_exception.assert_not_called()

        # Note: For real AIOProducer, the user callback is invoked directly by _handle_user_callback

    def test_flush_empty_buffer(self):
        """Test flushing empty buffer is no-op"""
        async def async_test():
            await self.batch_processor.flush_buffer()
            self.assertTrue(self.batch_processor.is_buffer_empty())

        self.loop.run_until_complete(async_test())

    def test_flush_buffer_with_messages(self):
        """Test successful buffer flush with messages"""
        async def async_test():
            future1 = self.loop.create_future()
            future2 = self.loop.create_future()
            msg1 = {'topic': 'topic1', 'value': 'test1'}
            msg2 = {'topic': 'topic1', 'value': 'test2'}

            self.batch_processor.add_message(msg1, future1)
            self.batch_processor.add_message(msg2, future2)

            success_future = self.loop.create_future()
            success_future.set_result("success")

            with patch.object(self.loop, 'run_in_executor', return_value=success_future):
                await self.batch_processor.flush_buffer()

            self.assertTrue(self.batch_processor.is_buffer_empty())

        self.loop.run_until_complete(async_test())

    def test_flush_buffer_selective_topic(self):
        """Test selective topic flushing"""
        async def async_test():
            future3 = self.loop.create_future()
            future4 = self.loop.create_future()
            msg3 = {'topic': 'topic1', 'value': 'test3'}
            msg4 = {'topic': 'topic2', 'value': 'test4'}

            self.batch_processor.add_message(msg3, future3)
            self.batch_processor.add_message(msg4, future4)

            success_future = self.loop.create_future()
            success_future.set_result("success")

            with patch.object(self.loop, 'run_in_executor', return_value=success_future):
                await self.batch_processor.flush_buffer(target_topic='topic1')

            self.assertEqual(self.batch_processor.get_buffer_size(), 1)

        self.loop.run_until_complete(async_test())

    def test_flush_buffer_exception_handling(self):
        """Test exception handling during buffer flush"""
        async def async_test():
            future = self.loop.create_future()
            msg = {'topic': 'topic1', 'value': 'test'}
            self.batch_processor.add_message(msg, future)

            exception = RuntimeError("Batch execution failed")

            with patch.object(self.loop, 'run_in_executor', side_effect=exception):
                with self.assertRaises(RuntimeError):
                    await self.batch_processor.flush_buffer()

            # Buffer should NOT be empty on exception - we want to retry
            self.assertFalse(self.batch_processor.is_buffer_empty())
            self.assertEqual(self.batch_processor.get_buffer_size(), 1)

        self.loop.run_until_complete(async_test())

    def test_kafka_executor_integration(self):
        """Test executing a batch operation via KafkaBatchExecutor"""
        async def async_test():
            batch_messages = [
                {'value': 'test1', 'callback': Mock()},
                {'value': 'test2', 'callback': Mock()},
            ]

            # Mock the executor to return a completed future
            future_result = self.loop.create_future()
            future_result.set_result("poll_result")

            with patch.object(self.loop, 'run_in_executor', return_value=future_result) as mock_run_in_executor:
                result = await self.kafka_executor.execute_batch('test-topic', batch_messages)

                # Verify run_in_executor was called
                mock_run_in_executor.assert_called_once()
                self.assertEqual(result, "poll_result")

        self.loop.run_until_complete(async_test())

    def _create_mixed_topic_messages(self):
        """Helper to create messages across multiple topics"""
        messages_data = []
        futures = []
        user_callbacks = []

        for i in range(4):
            future = Mock()
            future.done.return_value = False
            user_callback = Mock()
            msg_data = {
                'topic': f'topic{i % 2}',
                'value': f'unique_value_{i}',
                'key': f'unique_key_{i}',
                'user_callback': user_callback
            }

            self.batch_processor.add_message(msg_data, future)
            messages_data.append(msg_data)
            futures.append(future)
            user_callbacks.append(user_callback)

        return messages_data, futures, user_callbacks

    def _add_alternating_topic_messages(self):
        """Helper to add messages alternating between two topics"""
        futures = []
        for i in range(5):
            future = self.loop.create_future()
            msg_data = {
                'topic': f'topic{i % 2}',
                'value': f'test{i}',
                'key': f'key{i}'
            }
            self.batch_processor.add_message(msg_data, future)
            futures.append(future)
        return futures

    def test_batch_cycle_buffer_state(self):
        """Test buffer state during batch cycle"""
        self._add_alternating_topic_messages()
        self.assertEqual(self.batch_processor.get_buffer_size(), 5)
        self.assertFalse(self.batch_processor.is_buffer_empty())

    def test_batch_cycle_topic_grouping(self):
        """Test topic grouping in batch cycle"""
        self._add_alternating_topic_messages()
        topic_groups = self.batch_processor._group_messages_by_topic_and_partition()

        self.assertEqual(len(topic_groups), 2)
        self.assertIn(('topic0', -1), topic_groups)
        self.assertIn(('topic1', -1), topic_groups)
        self.assertEqual(len(topic_groups[('topic0', -1)]['messages']), 3)
        self.assertEqual(len(topic_groups[('topic1', -1)]['messages']), 2)

    def test_batch_cycle_message_preparation(self):
        """Test message preparation in batch cycle"""
        self._add_alternating_topic_messages()
        topic_groups = self.batch_processor._group_messages_by_topic_and_partition()

        batch_messages = self.batch_processor._prepare_batch_messages(
            topic_groups[('topic0', -1)]['messages']
        )

        self.assertEqual(len(batch_messages), 3)
        for batch_msg in batch_messages:
            self.assertNotIn('topic', batch_msg)
            self.assertIn('value', batch_msg)
            self.assertIn('key', batch_msg)

    def test_batch_message_preparation_with_mixed_sizes(self):
        """Test batch message preparation with mixed message sizes"""
        # Create test messages with different sizes
        messages = [
            {'topic': 'test-topic', 'value': 'small message'},
            {'topic': 'test-topic', 'value': 'x' * (5 * 1024 * 1024)},  # Large message
            {'topic': 'test-topic', 'value': 'another small'},
        ]
        futures = [asyncio.Future(), asyncio.Future(), asyncio.Future()]

        for msg, future in zip(messages, futures):
            self.batch_processor.add_message(msg, future)

        topic_groups = self.batch_processor._group_messages_by_topic_and_partition()
        topic_data = topic_groups[('test-topic', -1)]
        batch_messages = self.batch_processor._prepare_batch_messages(topic_data['messages'])

        self.assertEqual(len(batch_messages), 3)
        large_msg = next((msg for msg in batch_messages if len(str(msg.get('value', ''))) > 1000), None)
        self.assertIsNotNone(large_msg)

    def test_future_based_usage_pattern(self):
        """Test the recommended Future-based usage pattern instead of callbacks."""
        # Create test messages without user callbacks
        messages = [
            {'topic': 'test-topic', 'value': 'test1', 'key': 'key1'},
            {'topic': 'test-topic', 'value': 'test2', 'key': 'key2'},
        ]
        futures = [asyncio.Future(), asyncio.Future()]

        # Add messages to batch processor
        for msg, future in zip(messages, futures):
            self.batch_processor.add_message(msg, future)

        # Verify messages are in buffer
        self.assertEqual(self.batch_processor.get_buffer_size(), 2)

        # Simulate successful delivery by resolving futures
        mock_msg1 = Mock()
        mock_msg1.topic.return_value = 'test-topic'
        mock_msg1.value.return_value = b'test1'

        mock_msg2 = Mock()
        mock_msg2.topic.return_value = 'test-topic'
        mock_msg2.value.return_value = b'test2'

        # Applications should await these futures to get delivery results
        futures[0].set_result(mock_msg1)
        futures[1].set_result(mock_msg2)

        # Verify futures are resolved
        self.assertTrue(futures[0].done())
        self.assertTrue(futures[1].done())
        self.assertEqual(futures[0].result(), mock_msg1)
        self.assertEqual(futures[1].result(), mock_msg2)

    def test_future_based_error_handling(self):
        """Test Future-based error handling pattern."""
        # Create test message
        message = {'topic': 'test-topic', 'value': 'test', 'key': 'key'}
        future = asyncio.Future()

        # Add message to batch processor
        self.batch_processor.add_message(message, future)

        # Simulate delivery error by setting exception on future
        mock_error = RuntimeError("Delivery failed")
        future.set_exception(mock_error)

        # Verify future is resolved with exception
        self.assertTrue(future.done())
        with self.assertRaises(RuntimeError):
            future.result()

    def test_add_batches_back_to_buffer_basic(self):
        """Test adding batches back to buffer with basic message data"""
        from confluent_kafka.experimental.aio.producer._message_batch import create_message_batch

        # Create test futures
        future1 = asyncio.Future()
        future2 = asyncio.Future()

        # Create test batch with basic message data
        batch = create_message_batch(
            topic='test-topic',
            messages=[
                {'value': 'test1', 'key': 'key1'},
                {'value': 'test2', 'key': 'key2'}
            ],
            futures=[future1, future2],
            partition=0
        )

        # Ensure buffer is initially empty
        self.assertTrue(self.batch_processor.is_buffer_empty())

        # Add batch back to buffer
        self.batch_processor._add_batches_back_to_buffer([batch])

        # Verify buffer state
        self.assertEqual(self.batch_processor.get_buffer_size(), 2)
        self.assertFalse(self.batch_processor.is_buffer_empty())

        # Verify message data was reconstructed correctly
        self.assertEqual(self.batch_processor._message_buffer[0]['topic'], 'test-topic')
        self.assertEqual(self.batch_processor._message_buffer[0]['value'], 'test1')
        self.assertEqual(self.batch_processor._message_buffer[0]['key'], 'key1')

        self.assertEqual(self.batch_processor._message_buffer[1]['topic'], 'test-topic')
        self.assertEqual(self.batch_processor._message_buffer[1]['value'], 'test2')
        self.assertEqual(self.batch_processor._message_buffer[1]['key'], 'key2')

        # Verify futures are preserved
        self.assertEqual(self.batch_processor._buffer_futures[0], future1)
        self.assertEqual(self.batch_processor._buffer_futures[1], future2)

    def test_add_batches_back_to_buffer_empty_batch(self):
        """Test adding empty batch back to buffer"""
        from confluent_kafka.experimental.aio.producer._message_batch import create_message_batch

        # Create empty batch
        batch = create_message_batch(
            topic='test-topic',
            messages=[],
            futures=[],
            partition=0
        )

        initial_size = self.batch_processor.get_buffer_size()

        # Add empty batch back
        self.batch_processor._add_batches_back_to_buffer([batch])

        # Buffer size should remain unchanged
        self.assertEqual(self.batch_processor.get_buffer_size(), initial_size)


if __name__ == '__main__':
    # Run all tests
    unittest.main(verbosity=2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2025 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from confluent_kafka.cimpl import Producer
import inspect
import asyncio

from confluent_kafka.error import KeySerializationError, ValueSerializationError
from confluent_kafka.serialization import MessageField, SerializationContext

ASYNC_PRODUCER_POLL_INTERVAL: int = 0.2


class AsyncProducer(Producer):
    def __init__(
        self,
        conf: dict,
        loop: asyncio.AbstractEventLoop = None,
        poll_interval: int = ASYNC_PRODUCER_POLL_INTERVAL
    ):
        super().__init__(conf)

        self._loop = loop or asyncio.get_event_loop()
        self._poll_interval = poll_interval

        self._poll_task = None
        self._waiters: int = 0

    async def produce(
            self, topic, value=None, key=None, partition=-1,
            on_delivery=None, timestamp=0, headers=None
    ):
        fut = self._loop.create_future()
        self._waiters += 1
        try:
            if self._poll_task is None or self._poll_task.done():
                self._poll_task = asyncio.create_task(self._poll_dr(self._poll_interval))

            def wrapped_on_delivery(err, msg):
                if on_delivery is not None:
                    if inspect.iscoroutinefunction(on_delivery):
                        asyncio.run_coroutine_threadsafe(
                            on_delivery(err, msg),
                            self._loop
                        )
                    else:
                        self._loop.call_soon_threadsafe(on_delivery, err, msg)

                if err:
                    self._loop.call_soon_threadsafe(fut.set_exception, err)
                else:
                    self._loop.call_soon_threadsafe(fut.set_result, msg)

            super().produce(
                topic,
                value,
                key,
                headers=headers,
                partition=partition,
                timestamp=timestamp,
                on_delivery=wrapped_on_delivery
            )
            return await fut
        finally:
            self._waiters -= 1

    async def _poll_dr(self, interval: int):
        """Poll delivery reports at interval seconds"""
        while self._waiters:
            super().poll(0)
            await asyncio.sleep(interval)


class TestAsyncSerializingProducer(AsyncProducer):
    def __init__(self, conf):
        conf_copy = conf.copy()

        self._key_serializer = conf_copy.pop('key.serializer', None)
        self._value_serializer = conf_copy.pop('value.serializer', None)

        super(TestAsyncSerializingProducer, self).__init__(conf_copy)

    async def produce(
            self, topic, key=None, value=None, partition=-1,
            on_delivery=None, timestamp=0, headers=None):
        ctx = SerializationContext(topic, MessageField.KEY, headers)
        if self._key_serializer is not None:
            try:
                key = await self._key_serializer(key, ctx)
            except Exception as se:
                raise KeySerializationError(se)
        ctx.field = MessageField.VALUE
        if self._value_serializer is not None:
            try:
                value = await self._value_serializer(value, ctx)
            except Exception as se:
                raise ValueSerializationError(se)

        return await super().produce(
            topic, value, key,
            headers=headers,
            partition=partition,
            timestamp=timestamp,
            on_delivery=on_delivery
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from confluent_kafka import TopicPartition


def commit_and_check(consumer, topic, metadata):
    if metadata is None:
        consumer.commit(offsets=[TopicPartition(topic, 0, 1)], asynchronous=False)
    else:
        consumer.commit(offsets=[TopicPartition(topic, 0, 1, metadata)], asynchronous=False)

    offsets = consumer.committed([TopicPartition(topic, 0)], timeout=100)
    assert len(offsets) == 1
    assert offsets[0].metadata == metadata


def test_consumer_topicpartition_metadata(kafka_cluster):
    topic = kafka_cluster.create_topic_and_wait_propogation("test_topicpartition")
    consumer_conf = {'group.id': 'test_topicpartition'}

    c = kafka_cluster.consumer(consumer_conf)

    # Commit without any metadata.
    metadata = None
    commit_and_check(c, topic, metadata)

    # Commit with only ASCII metadata.
    metadata = 'hello world'
    commit_and_check(c, topic, metadata)

    # Commit with Unicode characters in metadata.
    metadata = 'नमस्ते दुनिया'
    commit_and_check(c, topic, metadata)

    # Commit with empty string as metadata.
    metadata = ''
    commit_and_check(c, topic, metadata)

    # Commit with invalid metadata (with null byte in the middle).
    metadata = 'xyz\x00abc'
    try:
        commit_and_check(c, topic, metadata)
        # We should never reach this point, since the prior statement should throw.
        assert False
    except ValueError as ve:
        assert 'embedded null character' in str(ve)

    c.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tests/integration/schema_registry/data/proto/NestedTestProto.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nBtests/integration/schema_registry/data/proto/NestedTestProto.proto\x12$tests.integration.serialization.data\x1a\x1fgoogle/protobuf/timestamp.proto\"\x8c\x01\n\x06UserId\x12\x17\n\rkafka_user_id\x18\x01 \x01(\tH\x00\x12\x17\n\rother_user_id\x18\x02 \x01(\x05H\x00\x12\x45\n\nanother_id\x18\x03 \x01(\x0b\x32/.tests.integration.serialization.data.MessageIdH\x00\x42\t\n\x07user_id\"\x17\n\tMessageId\x12\n\n\x02id\x18\x01 \x01(\t\"R\n\x0b\x43omplexType\x12\x10\n\x06one_id\x18\x01 \x01(\tH\x00\x12\x12\n\x08other_id\x18\x02 \x01(\x05H\x00\x12\x11\n\tis_active\x18\x03 \x01(\x08\x42\n\n\x08some_val\"\xd0\x04\n\rNestedMessage\x12=\n\x07user_id\x18\x01 \x01(\x0b\x32,.tests.integration.serialization.data.UserId\x12\x11\n\tis_active\x18\x02 \x01(\x08\x12\x1a\n\x12\x65xperiments_active\x18\x03 \x03(\t\x12<\n\x06status\x18\x05 \x01(\x0e\x32,.tests.integration.serialization.data.Status\x12G\n\x0c\x63omplex_type\x18\x06 \x01(\x0b\x32\x31.tests.integration.serialization.data.ComplexType\x12R\n\x08map_type\x18\x07 \x03(\x0b\x32@.tests.integration.serialization.data.NestedMessage.MapTypeEntry\x12O\n\x05inner\x18\x08 \x01(\x0b\x32@.tests.integration.serialization.data.NestedMessage.InnerMessage\x1a.\n\x0cMapTypeEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a/\n\x0cInnerMessage\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x0f\n\x03ids\x18\x02 \x03(\x05\x42\x02\x10\x01\"(\n\tInnerEnum\x12\x08\n\x04ZERO\x10\x00\x12\r\n\tALSO_ZERO\x10\x00\x1a\x02\x10\x01J\x04\x08\x0e\x10\x0fJ\x04\x08\x0f\x10\x10J\x04\x08\t\x10\x0cR\x03\x66ooR\x03\x62\x61r*\"\n\x06Status\x12\n\n\x06\x41\x43TIVE\x10\x00\x12\x0c\n\x08INACTIVE\x10\x01\x42\x41\n,io.confluent.kafka.serializers.protobuf.testB\x0fNestedTestProtoP\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tests.integration.schema_registry.data.proto.NestedTestProto_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n,io.confluent.kafka.serializers.protobuf.testB\017NestedTestProtoP\000'
  _NESTEDMESSAGE_MAPTYPEENTRY._options = None
  _NESTEDMESSAGE_MAPTYPEENTRY._serialized_options = b'8\001'
  _NESTEDMESSAGE_INNERMESSAGE.fields_by_name['ids']._options = None
  _NESTEDMESSAGE_INNERMESSAGE.fields_by_name['ids']._serialized_options = b'\020\001'
  _NESTEDMESSAGE_INNERENUM._options = None
  _NESTEDMESSAGE_INNERENUM._serialized_options = b'\020\001'
  _STATUS._serialized_start=988
  _STATUS._serialized_end=1022
  _USERID._serialized_start=142
  _USERID._serialized_end=282
  _MESSAGEID._serialized_start=284
  _MESSAGEID._serialized_end=307
  _COMPLEXTYPE._serialized_start=309
  _COMPLEXTYPE._serialized_end=391
  _NESTEDMESSAGE._serialized_start=394
  _NESTEDMESSAGE._serialized_end=986
  _NESTEDMESSAGE_MAPTYPEENTRY._serialized_start=821
  _NESTEDMESSAGE_MAPTYPEENTRY._serialized_end=867
  _NESTEDMESSAGE_INNERMESSAGE._serialized_start=869
  _NESTEDMESSAGE_INNERMESSAGE._serialized_end=916
  _NESTEDMESSAGE_INNERENUM._serialized_start=918
  _NESTEDMESSAGE_INNERENUM._serialized_end=958
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tests/integration/schema_registry/data/proto/PublicTestProto.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tests.integration.schema_registry.data.proto import TestProto_pb2 as tests_dot_integration_dot_schema__registry_dot_data_dot_proto_dot_TestProto__pb2

from tests.integration.schema_registry.data.proto.TestProto_pb2 import *

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nBtests/integration/schema_registry/data/proto/PublicTestProto.proto\x12$tests.integration.serialization.data\x1a<tests/integration/schema_registry/data/proto/TestProto.protoP\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tests.integration.schema_registry.data.proto.PublicTestProto_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python

from io import StringIO
import logging

import confluent_kafka
import confluent_kafka.avro
import confluent_kafka.admin

from tests.common import TestConsumer
from tests.common.schema_registry import TestAvroConsumer


class CountingFilter(logging.Filter):
    def __init__(self, name):
        self.name = name
        self.cnt = 0

    def filter(self, record):
        print(self.name, record.getMessage())
        self.cnt += 1
        print(record)


def _setup_string_buffer_logger(name):
    stringBuffer = StringIO()
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(stringBuffer)
    handler.setFormatter(logging.Formatter('%(name)s Logger | %(message)s'))
    logger.addHandler(handler)
    return stringBuffer, logger


def test_logging_consumer():
    """ Tests that logging works """

    logger = logging.getLogger('consumer')
    logger.setLevel(logging.DEBUG)
    f = CountingFilter('consumer')
    logger.addFilter(f)
    kc = TestConsumer({'group.id': 'test',
                       'debug': 'all'},
                      logger=logger)
    while f.cnt == 0:
        kc.poll(timeout=0.5)

    print('%s: %d log messages seen' % (f.name, f.cnt))

    kc.close()


def test_logging_avro_consumer():
    """ Tests that logging works """

    logger = logging.getLogger('avroconsumer')
    logger.setLevel(logging.DEBUG)
    f = CountingFilter('avroconsumer')
    logger.addFilter(f)

    kc = TestAvroConsumer({'schema.registry.url': 'http://example.com',
                           'group.id': 'test',
                           'debug': 'all'},
                          logger=logger)
    while f.cnt == 0:
        kc.poll(timeout=0.5)

    print('%s: %d log messages seen' % (f.name, f.cnt))

    kc.close()


def test_logging_producer():
    """ Tests that logging works """

    logger = logging.getLogger('producer')
    logger.setLevel(logging.DEBUG)
    f = CountingFilter('producer')
    logger.addFilter(f)

    p = confluent_kafka.Producer({'debug': 'all'}, logger=logger)

    while f.cnt == 0:
        p.poll(timeout=0.5)

    print('%s: %d log messages seen' % (f.name, f.cnt))


def test_logging_avro_producer():
    """ Tests that logging works """

    logger = logging.getLogger('avroproducer')
    logger.setLevel(logging.DEBUG)
    f = CountingFilter('avroproducer')
    logger.addFilter(f)

    p = confluent_kafka.avro.AvroProducer({'schema.registry.url': 'http://example.com',
                                           'debug': 'all'},
                                          logger=logger)

    while f.cnt == 0:
        p.poll(timeout=0.5)

    print('%s: %d log messages seen' % (f.name, f.cnt))


def test_logging_constructor():
    """ Verify different forms of constructors """

    for how in ['dict', 'dict+kwarg', 'kwarg']:
        logger = logging.getLogger('producer: ' + how)
        logger.setLevel(logging.DEBUG)
        f = CountingFilter(logger.name)
        logger.addFilter(f)

        if how == 'dict':
            p = confluent_kafka.Producer({'debug': 'all', 'logger': logger})
        elif how == 'dict+kwarg':
            p = confluent_kafka.Producer({'debug': 'all'}, logger=logger)
        elif how == 'kwarg':
            conf = {'debug': 'all', 'logger': logger}
            p = confluent_kafka.Producer(**conf)
        else:
            raise RuntimeError('Not reached')

        print('Test %s with %s' % (p, how))

        while f.cnt == 0:
            p.poll(timeout=0.5)

        print('%s: %s: %d log messages seen' % (how, f.name, f.cnt))


def test_producer_logger_logging_in_given_format():
    """Test that asserts that logging is working by matching part of the log message"""

    stringBuffer, logger = _setup_string_buffer_logger('Producer')

    p = confluent_kafka.Producer(
        {"bootstrap.servers": "test", "logger": logger, "debug": "msg"})
    val = 1
    while val > 0:
        val = p.flush()
    logMessage = stringBuffer.getvalue().strip()
    stringBuffer.close()
    print(logMessage)

    assert "Producer Logger | INIT" in logMessage


def test_consumer_logger_logging_in_given_format():
    """Test that asserts that logging is working by matching part of the log message"""

    stringBuffer, logger = _setup_string_buffer_logger('Consumer')

    c = TestConsumer(
        {"bootstrap.servers": "test", "group.id": "test", "logger": logger, "debug": "msg"})
    c.poll(0)

    logMessage = stringBuffer.getvalue().strip()
    stringBuffer.close()
    c.close()

    assert "Consumer Logger | INIT" in logMessage


def test_admin_logger_logging_in_given_format_when_provided_in_conf():
    """Test that asserts that logging is working by matching part of the log message"""

    stringBuffer, logger = _setup_string_buffer_logger('Admin')

    admin_client = confluent_kafka.admin.AdminClient(
        {"bootstrap.servers": "test", "logger": logger, "debug": "admin"})
    admin_client.poll(0)

    logMessage = stringBuffer.getvalue().strip()
    stringBuffer.close()

    assert "Admin Logger | INIT" in logMessage


def test_admin_logger_logging_in_given_format_when_provided_as_admin_client_argument():
    """Test that asserts that logging is working by matching part of the log message"""

    stringBuffer, logger = _setup_string_buffer_logger('Admin')

    admin_client = confluent_kafka.admin.AdminClient(
        {"bootstrap.servers": "test", "debug": "admin"}, logger=logger)
    admin_client.poll(0)

    logMessage = stringBuffer.getvalue().strip()
    stringBuffer.close()

    assert "Admin Logger | INIT" in logMessage
